import copy
import logging
import numpy as np
import matplotlib as mpl
from datetime import date, datetime, timezone
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import pyspedas
from fnmatch import filter as tname_filter
from time import sleep
from pyspedas.tplot_tools import tplot_wildcard_expand, tname_byindex, get_data, var_label_panel
from pyspedas.tplot_tools import lineplot, count_traces, makegap
from pyspedas.tplot_tools import specplot, specplot_make_1d_ybins, reduce_spec_dataset
from pyspedas.tplot_tools import get_var_label_ticks

# the following improves the x-axis ticks labels
import matplotlib.units as munits
import matplotlib.dates as mdates
converter = mdates.ConciseDateConverter()
munits.registry[np.datetime64] = converter
munits.registry[date] = converter
munits.registry[datetime] = converter

def pseudovar_component_props(varname: str):
    """ Return or calculate the plot properties for a single normal tplot variable


    Parameters
    ----------
    varname: str
        Name of the tplot variable to inspect

    Returns
    -------
    dict
        Dictionary of plot properties for this variable::
            is_spec: True if variable is a spectrogram
            ymin, ymax, ylog: Y axis limits
            zmin, zmax, zlog: Z axis limits (if variable is a spectrogram)

    """

    output_dict = {'ymin':np.nan, 'ymax':np.nan, 'ylog':False,
                   'zmin':np.nan, 'zmax': np.nan, 'zlog':False,
                   'is_spec': False}

    attrs = pyspedas.tplot_tools.data_quants[varname].attrs
    plot_opts = attrs.get('plot_options')
    if plot_opts is not None:
        yaxis_opt = plot_opts.get('yaxis_opt')
        if yaxis_opt is not None:
            ylog = yaxis_opt['y_axis_type']
            yrange = yaxis_opt.get('y_range')
    else:
        ylog = False
        yrange=[None, None]

    plot_extras = attrs['plot_options']['extras']
    dat = get_data(varname)

    if ylog is None or ylog is False or ylog == '' or ylog.lower() == 'linear':
        output_dict['ylog'] = False
        ylog = False
    else:
        output_dict['ylog'] = True
        ylog = True

    # ylog is now guaranteed to be boolean

    if plot_extras.get('spec') is not None:
        spec = plot_extras['spec']
        if spec:
            # Inspect a specplot variable
            output_dict['is_spec']=True
            if yrange is not None:
                output_dict['ymin'] = yrange[0]
                output_dict['ymax'] = yrange[1]
            else:
                # Figure out which attribute to use for bin centers (copied from specplot, maybe
                # should be pulled out into a separate routine?)
                var_data=dat
                if len(var_data) == 3:
                    if hasattr(var_data, 'v'):
                        input_bin_centers = var_data.v
                    elif hasattr(var_data, 'v1'):
                        input_bin_centers = var_data.v1
                    else:
                        logging.warning("Multidimensional variable %s has no v or v1 attribute", varname)
                elif len(var_data) == 4:
                    if hasattr(var_data, 'v1'):
                        if 'spec_dim_to_plot' in plot_extras:
                            if plot_extras['spec_dim_to_plot'] == 'v1':
                                input_bin_centers = var_data.v1
                    if hasattr(var_data, 'v2'):
                        if 'spec_dim_to_plot' in plot_extras:
                            if plot_extras['spec_dim_to_plot'] == 'v2':
                                input_bin_centers = var_data.v2

                output_bin_boundaries = specplot_make_1d_ybins(None, input_bin_centers, ylog, no_regrid=True)
                output_dict['ymin']=np.nanmin(output_bin_boundaries)
                output_dict['ymax']=np.nanmax(output_bin_boundaries)

            zaxis_options = attrs['plot_options']['zaxis_opt']
            zrange = zaxis_options.get('zrange')
            if zrange is not None:
                output_dict['zmin'] = zrange[0]
                output_dict['zmax'] = zrange[1]
            else:
                # Determine Z range from data array
                output_dict['zmin'] = np.nanmin(dat.y)
                output_dict['zmax'] = np.nanmax(dat.y)
            zlog = zaxis_options.get('z_axis_type')
            if zlog is None or zlog.lower() == 'linear':
                output_dict['zlog'] = False
            else:
                output_dict['zlog'] = True


    else:
        # Inspect a line plot variable
        output_dict['is_spec']=False
        if yrange is not None:
            output_dict['ymin'] = yrange[0]
            output_dict['ymax'] = yrange[1]
        else:
            output_dict['ymin'] = np.nanmin(dat.y)
            output_dict['ymax'] = np.nanmax(dat.y)
    return output_dict

def gather_pseudovar_props(pseudovars:list[str]):
    """ Inspect the components of a pseudovariable to determine plot limits and other settings

    If any component has y or z ranges/scales set, they will be used, otherwise limits will be determined from
    the data values.

    If any component specifies log scaling, it will be applied to all components, otherwise linear scaling will
    be used for all

    Parameters
    ----------
    pseudovars: list[str]
        List of tplot composite variable components to inspect

    Returns
    -------
    dict
        A dictionary containing the plot properties to apply to each component variable::
            line_ymin, line_ymax, line_ylog : Y axis limits for line plots
            spec_ymin, spec_ymax, spec_ylog : Y axis limits for spectrograms
            spec_zmin, spec_zmax, spec_zlog : Z axis limits for spectrograms
            has_line_plots: True if any component is a line plot
            has_spec_plots: True if any component is a spectrogram plot

    """
    output_dict = { 'has_line_plots': False,
                    'has_spec_plots': False,
                    'line_ymin': np.nan,
                    'line_ymax': np.nan,
                    'line_ylog': False,
                    'spec_ymin': np.nan,
                    'spec_ymax': np.nan,
                    'spec_ylog': False,
                    'spec_zmin': np.nan,
                    'spec_zmax': np.nan,
                    'spec_zlog': False,
                    'first_spec_var': None}


    for var in pseudovars:
        props = pseudovar_component_props(var)
        #print(f"{var}: is_spec: {props['is_spec']} ymin: {props['ymin']} ymax: {props['ymax']} ylog: {props['ylog']} zmin: {props['zmin']} zmax: {props['zmax']} zlog: {props['zlog']}")
        if props['is_spec']:
            output_dict['has_spec_plots'] = True
            if output_dict['first_spec_var'] is None:
                output_dict['first_spec_var'] = var
            if props['zlog']:
                output_dict['spec_zlog'] = True
            if props['ylog']:
                output_dict['spec_ylog'] = True
            output_dict['spec_zmin']= np.nanmin([output_dict['spec_zmin'], props['zmin']])
            output_dict['spec_zmax']= np.nanmax([output_dict['spec_zmax'], props['zmax']])
            output_dict['spec_ymin']= np.nanmin([output_dict['spec_ymin'], props['ymin']])
            output_dict['spec_ymax']= np.nanmax([output_dict['spec_ymax'], props['ymax']])
        else:
            output_dict['has_line_plots'] = True
            if props['ylog']:
                output_dict['line_ylog'] = True
            output_dict['line_ymin'] = np.nanmin([output_dict['line_ymin'], props['ymin']])
            output_dict['line_ymax'] = np.nanmax([output_dict['line_ymax'], props['ymax']])
    return output_dict


def tplot(variables,
          trange=None,
          var_label=None,
          xsize=None,
          ysize=None,
          save_png='',
          save_eps='',
          save_svg='',
          save_pdf='',
          save_jpeg='',
          dpi=None,
          display=True,
          fig=None,
          axis=None,
          running_trace_count=None,
          pseudo_idx=None,
          pseudo_right_axis=False,
          pseudo_xaxis_options=None,
          pseudo_yaxis_options=None,
          pseudo_zaxis_options=None,
          pseudo_line_options=None,
          pseudo_extra_options=None,
          show_colorbar=True,
          slice=False,
          return_plot_objects=False):
    """
    Plot tplot variables to the display, or as saved files, using Matplotlib

    Parameters
    ----------
        variables: str or list of str, required
            List of tplot variables to be plotted.  Space-delimited strings may be used instead f
            lists.  Wildcards will be expanded.
        trange: list of string or float, optional
            If set, this time range will be used, temporarily overriding any previous xlim or timespan calls
        var_label : str or list of str, optional
            A list of variables to be displayed as values underneath the X axis major tick marks
        xsize: float, optional
            Plot size in the horizontal direction (in inches)
        ysize: float, optional
            Plot size in the vertical direction (in inches)
        dpi: float, optional
            The resolution of the plot in dots per inch
        save_png : str, optional
            A full file name and path.
            If this option is set, the plot will be automatically saved to the file name provided in PNG format.
        save_eps : str, optional
            A full file name and path.
            If this option is set, the plot will be automatically saved to the file name provided in EPS format.
        save_jpeg : str, optional
            A full file name and path.
            If this option is set, the plot will be automatically saved to the file name provided in JPEG format.
        save_pdf : str, optional
            A full file name and path.
            If this option is set, the plot will be automatically saved to the file name provided in PDF format.
        save_svg : str, optional
            A full file name and path.
            If this option is set, the plot will be automatically saved to the file name provided in SVG format.
        dpi: float, optional
            The resolution of the plot in dots per inch
        display: bool, optional
            If True, then this function will display the plotted tplot variables. Use False to suppress display (for example, if
            saving to a file, or returning plot objects to be displayed later). Default: True
        fig: Matplotlib figure object
            Use an existing figure to plot in (mainly for recursive calls to render composite variables)
        axis: Matplotlib axes object
            Use an existing set of axes to plot on (mainly for recursive calls to render composite variables)
        running_trace_count: int, optional
            In recursive calls for rendering composite variables, the index of the trace currently being rendered.
        pseudo_idx: int, optional
            In recursive calls for rendering composite variables, the index of the variable component currently being rendered.
        pseudo_right_axis: bool, optional
            In recursive calls for rendering composite variables, a flag to indicate the Y scale should be placed on
            the right Y axis.
        pseudo_xaxis_options: dict, optional
            In recursice calls for rendering composite variables, X axis options inherited from the parent variable
        pseudo_yaxis_options: dict, optional
            In recursive calls for rendering composite variables, Y axis options inherited from the parent variable
        pseudo_zaxis_options: dict, optional
            In recursive calls for rendering composite variables, Z axis options inherited from the parent variable
        pseudo_line_options: dict, optional
            In recursive calls for rendering composite variables, line options inherited from the parent variable
        pseudo_extra_options: dict, optional
            In recursive calls for rendering composite variables, extra options inherited from the parent variable
        show_colorbar: bool, optional
            Show a colorbar showing the Z scale for spectrogram plots.
        slice: bool, optional
            If True, show an interactive window with a plot of Z versus Y values for the X axis (time) value under the cursor. Default: False
        return_plot_objects: bool, optional
            If true, returns the matplotlib fig and axes objects for further manipulation. Default: False

    Returns
    -------
        Any
            Returns matplotlib fig and axes objects, if return_plot_objects==True

    Examples
    --------
        >>> # Plot a single line plot
        >>> import pyspedas
        >>> x_data = [2,3,4,5,6]
        >>> y_data = [1,2,3,4,5]
        >>> pyspedas.store_data("Variable1", data={'x':x_data, 'y':y_data})
        >>> pyspedas.tplot("Variable1")

        >>> # Plot two variables
        >>> x_data = [1,2,3,4,5]
        >>> y_data = [[1,5],[2,4],[3,3],[4,2],[5,1]]
        >>> pyspedas.store_data("Variable2", data={'x':x_data, 'y':y_data})
        >>> pyspedas.tplot(["Variable1", "Variable2"])

        >>> # Plot two plots, adding a third variable's values as annotations at X axis major tick marks
        >>> x_data = [1,2,3]
        >>> y_data = [ [1,2,3] , [4,5,6], [7,8,9] ]
        >>> v_data = [1,2,3]
        >>> pyspedas.store_data("Variable3", data={'x':x_data, 'y':y_data, 'v':v_data})
        >>> pyspedas.options("Variable3", 'spec', 1)
        >>> pyspedas.tplot(["Variable2", "Variable3"], var_label='Variable1')

    """
    # This call resolves wildcard patterns and converts integers to variable names
    variables = tplot_wildcard_expand(variables)
    if len(variables) == 0:
        logging.warning("tplot: No matching tplot names were found")
        return


    varlabel_style = pyspedas.tplot_tools.tplot_opt_glob.get('varlabel_style')
    if varlabel_style is None or varlabel_style.lower() == 'extra_axes':
        num_panels = len(variables)
        panel_sizes = [1]*num_panels
    else: # varlabel_style 'extra_panel'
        num_panels = len(variables) + 1
        panel_sizes = [1] * len(variables) + [0.1 * (len(var_label) + 2)]

    # support for the panel_size option
    for var_idx, variable in enumerate(variables):
        if pyspedas.tplot_tools.data_quants.get(variable) is None:
            continue
        panel_size = pyspedas.tplot_tools.data_quants[variable].attrs['plot_options']['extras'].get('panel_size')
        if panel_size is not None:
            panel_sizes[var_idx] = panel_size

    if xsize is None:
        xsize = pyspedas.tplot_tools.tplot_opt_glob.get('xsize')
        if xsize is None:
            xsize = 12

    if ysize is None:
        ysize = pyspedas.tplot_tools.tplot_opt_glob.get('ysize')
        if ysize is None:
            if num_panels > 4:
                ysize = 8
            else:
                # This was previously set to 5, which resulted in a non-monotonic progression of panel heights
                default_y_sizes = [5, 5, 6, 7, 8]
                ysize = default_y_sizes[num_panels]

    # The logic here for handling the 'right_axis' option is pretty convoluted, and makes a number of assumptions
    # that may not be warranted: mainly that right_axis will only be set on pseudovariables, and that if
    # set, the first sub-variable gets the left axis and all other sub-variables get a newly twinx-ed right axis.
    # "right axis" implies there will only be at most two Y axes. What if more scales are needed?
    # There also seems to be some conflation of the set of axes for the whole stack of plots, versus
    # the single (left or possibly right) axis for the variable currently being rendered.
    #
    # This whole concept is kind of a mess at the moment.  For now, we'll make it work for the
    # most likely use case, plotting a single spectrum variable followed by a single line variable
    # (for example, an energy spectrum plus spacecraft potential).  JWL 2024-03-26

    if fig is None and axis is None:
        fig, axes = plt.subplots(nrows=num_panels, sharex=True, gridspec_kw={'height_ratios': panel_sizes}, layout='constrained')
        fig.set_size_inches(xsize, ysize)
        plot_title = pyspedas.tplot_tools.tplot_opt_glob['title_text']
        fig.suptitle(plot_title)
        if (plot_title is not None) and plot_title != '':
            if 'title_size' in pyspedas.tplot_tools.tplot_opt_glob:
                title_size = pyspedas.tplot_tools.tplot_opt_glob['title_size']
                fig.suptitle(plot_title, fontsize=title_size)
            else:
                fig.suptitle(plot_title)
        # support for matplotlib styles
        style = pyspedas.tplot_tools.tplot_opt_glob.get('style')
        if style is not None:
            plt.style.use(style)
    else:
        # fig and axis have been passed as parameters, most likely a recursive tplot call to render
        # a pseudovariable
        if pseudo_idx == 0 or pseudo_right_axis == False:
            # setting up first axis
            axes = axis
        elif pseudo_idx > 0 and pseudo_right_axis:
            # generate and use the right axis?  probably still wrong...
            axes = axis.twinx()
        # support for matplotlib styles -- shouldn't be needed if processing pseudovar components?
        style = pyspedas.tplot_tools.tplot_opt_glob.get('style')
        if style is not None:
            plt.style.use(style)


    axis_font_size = pyspedas.tplot_tools.tplot_opt_glob.get('axis_font_size')

    colorbars = {}

    for idx, variable in enumerate(variables):
        var_data_org = get_data(variable, dt=True)
        var_metadata = get_data(variable, metadata=True)

        #Check for a 3d variable, call reduce_spec_dataset
        if hasattr(var_data_org, 'v1') and hasattr(var_data_org, 'v2'):
            temp_dq = reduce_spec_dataset(name=variable)
            var_data_org = get_data(variable, dt=True, data_quant_in=temp_dq)
        
        if var_data_org is None:
            logging.info('Variable not found: ' + variable)
            continue

        var_data = copy.deepcopy(var_data_org)

        # plt.subplots returns a list of axes for multiple panels
        # but only a single axis for a single panel
        if num_panels == 1:
            this_axis = axes
        else:
            this_axis = axes[idx]

        # we need to track the variable name in the axis object
        # for spectrogram slices
        this_axis.var_name = variable

        pseudo_var = False
        overplots = None
        spec = False

        var_quants = pyspedas.tplot_tools.data_quants[variable]

        if not isinstance(var_quants, dict):
            overplots = var_quants.attrs['plot_options'].get('overplots_mpl')
            if overplots is not None and len(overplots) > 0:
                pseudo_var = True

        # deal with pseudo-variables first
        if isinstance(var_data, list) or isinstance(var_data, str) or pseudo_var:
            # this is a pseudo variable
            if isinstance(var_data, str):
                var_data = var_data.split(' ')

            if pseudo_var:
                pseudo_vars = overplots
            else:
                pseudo_vars = var_data

            # pseudo variable metadata should override the metadata
            # for individual variables
            xaxis_options = None
            yaxis_options = None
            zaxis_options = None
            line_opts = None
            plot_extras = None
            if pseudo_var:
                plot_extras = var_quants.attrs['plot_options']['extras']
                if plot_extras.get('spec') is not None:
                    spec = True

                if plot_extras.get('right_axis') is not None:
                    if plot_extras.get('right_axis'):
                        pseudo_right_axis = True

                if pseudo_right_axis or spec:
                    plot_extras = None
                else:
                    yaxis_options = var_quants.attrs['plot_options']['yaxis_opt']
                    zaxis_options = var_quants.attrs['plot_options']['zaxis_opt']
                    line_opts = var_quants.attrs['plot_options']['line_opt']

            traces_processed = 0
            pseudovar_props = gather_pseudovar_props(pseudo_vars)
            #
            # Determine which Y axis options need to be changed to accommodate multiple component variables
            # There might be line variables and spectra, each with their own overall yscale and yrange.
            # If both are present, the spectra will take priority.
            # If explicit y or z scales or ranges are set on the pseudovariable, do not override.
            if pseudovar_props['has_spec_plots']:
                if pseudovar_props['spec_ylog']:
                    new_yscale = 'log'
                else:
                    new_yscale = 'linear'
                new_yr = [pseudovar_props['spec_ymin'], pseudovar_props['spec_ymax']]
            else:
                if pseudovar_props['line_ylog']:
                    new_yscale = 'log'
                else:
                    new_yscale = 'linear'
                new_yr = [pseudovar_props['line_ymin'], pseudovar_props['line_ymax']]
            override_yopts = {}
            if pseudo_right_axis:
                # Who knows?  Let the component variable Y-axis fight it out...
                override_yopts = {}
            elif yaxis_options is None:
                override_yopts = {'y_range':new_yr, 'y_range_user':True,'y_axis_style':new_yscale}
            else:
                if yaxis_options.get('y_range') is None:
                    override_yopts['y_range'] = new_yr
                    override_yopts['y_range_user'] = True
                if yaxis_options.get('y_axis_style') is None:
                    override_yopts['y_axis_style'] = new_yscale
            if yaxis_options is not None:
                yaxis_options = yaxis_options | override_yopts
            else:
                yaxis_options = override_yopts

            # Determine which Z axis options need to be changed to accommodate multiple component variables
            #
            # We only care about specplots here.
            new_zr = [ None, pseudovar_props['spec_zmax']]
            if pseudovar_props['spec_zlog']:
                new_zscale='log'
            else:
                new_zscale='linear'
            override_zopts = {}
            if zaxis_options is None:
                override_zopts = {'z_range':new_zr, 'z_axis_style':new_zscale}
            else:
                if zaxis_options.get('z_range') is None:
                    override_zopts['z_range'] = new_zr
                if zaxis_options.get('z_axis_style') is None:
                    override_zopts['z_axis_style'] = new_zscale
            if zaxis_options is not None:
                zaxis_options = zaxis_options | override_zopts
            else:
                zaxis_options = override_zopts

            for pseudo_idx, var in enumerate(pseudo_vars):
                # We're plotting a pseudovariable.  Iterate over the sub-variables, keeping track of how many
                # traces have been plotted so far, so we can correctly match option values to traces. The pseudovariable
                # y_axis, z_axis, line and extra options are passed as parameters so they can be merged with the
                # sub-variable options, with any pseudovar options overriding the sub-variable options.
                trace_count_thisvar = count_traces(var)
                if var == pseudovar_props['first_spec_var']:
                    pseudo_show_colorbar=True
                else:
                    pseudo_show_colorbar=False
                tplot(var,
                      trange=trange,
                      return_plot_objects=return_plot_objects,
                      xsize=xsize, ysize=ysize,
                      fig=fig, axis=this_axis, display=False,
                      running_trace_count=traces_processed,
                      pseudo_idx=pseudo_idx,
                      pseudo_xaxis_options=xaxis_options, pseudo_yaxis_options=yaxis_options, pseudo_zaxis_options=zaxis_options,
                      pseudo_line_options=line_opts, pseudo_extra_options=plot_extras,
                      pseudo_right_axis=pseudo_right_axis,
                      show_colorbar=pseudo_show_colorbar)
                traces_processed += trace_count_thisvar
            

            continue


        #if data_gap is an option for this variable, or if it's a add
        #gaps here; an individual gap setting should override the
        #global setting
        plot_extras = var_quants.attrs['plot_options']['extras']
        if plot_extras.get('data_gap') is not None and plot_extras.get('data_gap') > 0:
            var_data = makegap(var_data, dt = plot_extras.get('data_gap'))
        else:
            if pyspedas.tplot_tools.tplot_opt_glob['data_gap'] is not None and pyspedas.tplot_tools.tplot_opt_glob['data_gap'] > 0:
                var_data = makegap(var_data, dt = pyspedas.tplot_tools.tplot_opt_glob['data_gap'])

        # set the x-axis range, if it was set with xlim or tlimit or the trange parameter
        if trange is None and pyspedas.tplot_tools.tplot_opt_glob.get('x_range') is None:
            var_data_times = var_data.times
            time_idxs = np.arange(len(var_data_times))
        else:
            if trange is not None:
                if len(trange) != 2:
                    logging.error('Invalid trange setting: must be a 2-element list or array')
                    return
                if isinstance(trange[0], str):
                    x_range = pyspedas.tplot_tools.time_double(trange) # seconds since epoch
                    x_range_start = x_range[0]
                    x_range_stop = x_range[1]
            else:
                x_range = pyspedas.tplot_tools.tplot_opt_glob['x_range']  # Seconds since epoch
                x_range_start = x_range[0]
                x_range_stop = x_range[1]

            # Check for NaN or inf in x_range
            if not np.isfinite(x_range_start):
                logging.warning('tplot: x_range start is not finite, replacing with 0')
                x_range_start = 0

            if not np.isfinite(x_range_stop):
                logging.warning('tplot: x_range end is not finite, replacing with 0')
                x_range_stop = 0

            # Convert to np.datetime64 with nanosecond precision
            x_range = np.array(np.array([x_range_start*1e9, x_range_stop*1e9]),dtype='datetime64[ns]')
            this_axis.set_xlim(x_range)
            time_idxs = np.argwhere((var_data.times >= x_range[0]) & (var_data.times <= x_range[1])).flatten()
            if len(time_idxs) == 0:
                logging.info('No data found in the time range: ' + variable)
                continue
            var_data_times = var_data.times[time_idxs]

        var_times = var_data_times

        # set some more plot options
        xaxis_options = var_quants.attrs['plot_options']['xaxis_opt']
        if pseudo_xaxis_options is not None and len(pseudo_xaxis_options) > 0:
            merged_xaxis_options = xaxis_options | pseudo_xaxis_options
            xaxis_options = merged_xaxis_options

        yaxis_options = var_quants.attrs['plot_options']['yaxis_opt']
        if pseudo_yaxis_options is not None and len(pseudo_yaxis_options) > 0:
            merged_yaxis_options = yaxis_options | pseudo_yaxis_options
            yaxis_options = merged_yaxis_options

        zaxis_options = var_quants.attrs['plot_options']['zaxis_opt']
        if pseudo_zaxis_options is not None and len(pseudo_zaxis_options) > 0:
            merged_zaxis_options = zaxis_options | pseudo_zaxis_options
            zaxis_options = merged_zaxis_options

        line_opts = var_quants.attrs['plot_options']['line_opt']
        if pseudo_line_options is not None and len(pseudo_line_options) > 0:
            merged_line_opts = line_opts | pseudo_line_options
            line_opts = merged_line_opts

        if line_opts is not None:
            if 'name' in line_opts:
                this_axis.set_title(line_opts['name'])
            elif 'title' in line_opts:
                this_axis.set_title(line_opts['title'])

        plot_extras = var_quants.attrs['plot_options']['extras']
        if pseudo_extra_options is not None and len(pseudo_extra_options) > 0:
            merged_plot_extras = plot_extras | pseudo_extra_options
            plot_extras = merged_plot_extras

        xtitle = None
        if xaxis_options.get('axis_label') is not None:
            xtitle = xaxis_options['axis_label']


        xsubtitle = ''
        if xaxis_options.get('axis_subtitle') is not None:
            xsubtitle = xaxis_options['axis_subtitle']

        if style is None:
            xtitle_color = 'black'
        else:
            xtitle_color = None

        if xaxis_options.get('axis_color') is not None:
            xtitle_color = xaxis_options['axis_color']

        ylog = yaxis_options['y_axis_type']

        if ylog == 'log':
            this_axis.set_yscale('log')
        else:
            this_axis.set_yscale('linear')

        ytitle = yaxis_options['axis_label']
        if ytitle == '':
            ytitle = variable

        ysubtitle = ''
        if yaxis_options.get('axis_subtitle') is not None:
            ysubtitle = yaxis_options['axis_subtitle']

        # replace some common superscripts
        ysubtitle = replace_common_exp(ysubtitle)

        if axis_font_size is not None:
            this_axis.tick_params(axis='x', labelsize=axis_font_size)
            this_axis.tick_params(axis='y', labelsize=axis_font_size)

        char_size = pyspedas.tplot_tools.tplot_opt_glob.get('charsize')
        if char_size is None:
            char_size = 12

        if plot_extras.get('char_size') is not None:
            char_size = plot_extras['char_size']

        user_set_yrange = yaxis_options.get('y_range_user')
        if user_set_yrange is not None:
            # the user has set the yrange manually
            yrange = yaxis_options['y_range']
            if not np.isfinite(yrange[0]):
                yrange[0] = None
            if not np.isfinite(yrange[1]):
                yrange[1] = None
            this_axis.set_ylim(yrange)

        ymajor_ticks = yaxis_options.get('y_major_ticks')
        if ymajor_ticks is not None:
            this_axis.set_yticks(ymajor_ticks)

        yminor_tick_interval = yaxis_options.get('y_minor_tick_interval')
        if yminor_tick_interval is not None and ylog != 'log':
            this_axis.yaxis.set_minor_locator(plt.MultipleLocator(yminor_tick_interval))

        if style is None:
            ytitle_color = 'black'
        else:
            ytitle_color = None

        if yaxis_options.get('axis_color') is not None:
            ytitle_color = yaxis_options['axis_color']

        if xtitle is not None and xtitle != '':
            if xtitle_color is not None:
                this_axis.set_xlabel(xtitle + '\n' + xsubtitle, fontsize=char_size, color=xtitle_color)
            else:
                this_axis.set_xlabel(xtitle + '\n' + xsubtitle, fontsize=char_size)

        if ytitle_color is not None:
            this_axis.set_ylabel(ytitle + '\n' + ysubtitle, fontsize=char_size, color=ytitle_color)
        else:
            this_axis.set_ylabel(ytitle + '\n' + ysubtitle, fontsize=char_size)

        border = True
        if plot_extras.get('border') is not None:
            border = plot_extras['border']

        if border == False:
            this_axis.axis('off')

        # axis tick options
        if plot_extras.get('xtickcolor') is not None:
            this_axis.tick_params(axis='x', color=plot_extras.get('xtickcolor'))

        if plot_extras.get('ytickcolor') is not None:
            this_axis.tick_params(axis='y', color=plot_extras.get('ytickcolor'))

        if plot_extras.get('xtick_direction') is not None:
            this_axis.tick_params(axis='x', direction=plot_extras.get('xtick_direction'))

        if plot_extras.get('ytick_direction') is not None:
            this_axis.tick_params(axis='y', direction=plot_extras.get('ytick_direction'))

        if plot_extras.get('xtick_length') is not None:
            this_axis.tick_params(axis='x', length=plot_extras.get('xtick_length'))

        if plot_extras.get('ytick_length') is not None:
            this_axis.tick_params(axis='y', length=plot_extras.get('ytick_length'))

        if plot_extras.get('xtick_width') is not None:
            this_axis.tick_params(axis='x', width=plot_extras.get('xtick_width'))

        if plot_extras.get('ytick_width') is not None:
            this_axis.tick_params(axis='y', width=plot_extras.get('ytick_width'))

        if plot_extras.get('xtick_labelcolor') is not None:
            this_axis.tick_params(axis='x', labelcolor=plot_extras.get('xtick_labelcolor'))

        if plot_extras.get('ytick_labelcolor') is not None:
            this_axis.tick_params(axis='y', labelcolor=plot_extras.get('ytick_labelcolor'))

        # determine if this is a line plot or a spectrogram
        spec = False
        if plot_extras.get('spec') is not None:
            spec = plot_extras['spec']

        if spec:
            # create spectrogram plots
            plot_created = specplot(var_data, var_times, this_axis, yaxis_options, zaxis_options, plot_extras, colorbars, axis_font_size, fig, variable, time_idxs=time_idxs, style=style)
            if not plot_created:
                continue
        else:
            # create line plots
            plot_created = lineplot(var_data, var_times, this_axis, line_opts, yaxis_options, plot_extras, running_trace_count=running_trace_count, time_idxs=time_idxs, style=style, var_metadata=var_metadata)
            if not plot_created:
                continue

        # apply any vertical/horizontal bars
        if pyspedas.tplot_tools.data_quants[variable].attrs['plot_options'].get('time_bar') is not None:
            time_bars = pyspedas.tplot_tools.data_quants[variable].attrs['plot_options']['time_bar']

            for time_bar in time_bars:
                # vertical bars
                if time_bar['dimension'] == 'height':
                    this_axis.axvline(x=datetime.fromtimestamp(time_bar['location'], tz=timezone.utc),
                        color=np.array(time_bar.get('line_color'))/256.0, lw=time_bar.get('line_width'),
                                      linestyle=time_bar.get('line_dash'))

                # horizontal bars
                if time_bar['dimension'] == 'width':
                    this_axis.axhline(y=time_bar['location'], color=np.array(time_bar.get('line_color'))/256.0,
                                      lw=time_bar.get('line_width'),
                                      linestyle=time_bar.get('line_dash'))

        # highlight time intervals
        if pyspedas.tplot_tools.data_quants[variable].attrs['plot_options'].get('highlight_intervals') is not None:
            highlight_intervals = pyspedas.tplot_tools.data_quants[variable].attrs['plot_options']['highlight_intervals']

            for highlight_interval in highlight_intervals:
                hightlight_opts = copy.deepcopy(highlight_interval)
                del hightlight_opts['location']
                if highlight_interval['edgecolor'] is not None or highlight_interval['facecolor'] is not None:
                    del hightlight_opts['color']

                this_axis.axvspan(mdates.date2num(datetime.fromtimestamp(highlight_interval['location'][0], timezone.utc)),
                                  mdates.date2num(datetime.fromtimestamp(highlight_interval['location'][1], timezone.utc)),
                                  **hightlight_opts)

        # add annotations
        if pyspedas.tplot_tools.data_quants[variable].attrs['plot_options']['extras'].get('annotations') is not None:
            annotations = pyspedas.tplot_tools.data_quants[variable].attrs['plot_options']['extras']['annotations']
            for annotation in annotations:
                this_axis.annotate(annotation['text'], annotation['position'],
                                   xycoords=annotation['xycoords'],
                                   fontsize=annotation['fontsize'],
                                   alpha=annotation['alpha'],
                                   fontfamily=annotation['fontfamily'],
                                   fontvariant=annotation['fontvariant'],
                                   fontstyle=annotation['fontstyle'],
                                   fontstretch=annotation['fontstretch'],
                                   fontweight=annotation['fontweight'],
                                   rotation=annotation['rotation'],
                                   color=annotation['color'])

    # apply any addition x-axes (or panel) specified by the var_label keyword
    if var_label is not None:
        if varlabel_style is None or varlabel_style.lower() == 'extra_axes':
            varlabels_extra_axes(num_panels, this_axis, var_label)
        else:
            var_label_panel(variables, var_label, axes,  axis_font_size)

    # add the color bars to any spectra
    for idx, variable in enumerate(variables):
        if pyspedas.tplot_tools.data_quants.get(variable) is None:
            continue
        plot_extras = pyspedas.tplot_tools.data_quants[variable].attrs['plot_options']['extras']
        zaxis_options = pyspedas.tplot_tools.data_quants[variable].attrs['plot_options']['zaxis_opt']
        if plot_extras.get('spec') is not None:
            spec = plot_extras['spec']
        else:
            spec = False


        if spec and show_colorbar:
            if colorbars.get(variable) is None:
                continue

            if num_panels == 1:
                this_axis = axes
            else:
                this_axis = axes[idx]

            colorbar = fig.colorbar(colorbars[variable]['im'], ax=this_axis)

            if style is None:
                ztitle_color = 'black'
            else:
                ztitle_color = None

            if zaxis_options is None:
                continue

            if zaxis_options.get('axis_color') is not None:
                ztitle_color = zaxis_options['axis_color']

            ztitle_text = colorbars[variable]['ztitle']
            zsubtitle_text = colorbars[variable]['zsubtitle']

            # replace some common superscripts
            ztitle_text = replace_common_exp(ztitle_text)
            zsubtitle_text = replace_common_exp(zsubtitle_text)

            if ztitle_color is not None:
                colorbar.set_label(ztitle_text + '\n ' + zsubtitle_text,
                                   color=ztitle_color, fontsize=char_size)
            else:
                colorbar.set_label(ztitle_text + '\n ' + zsubtitle_text,
                                   fontsize=char_size)

    # plt.tight_layout()
    fig.canvas.draw()

    if save_png is not None and save_png != '':
        if not save_png.endswith('.png'):
            save_png += '.png'
        plt.savefig(save_png, dpi=dpi)

    if save_eps is not None and save_eps != '':
        if not save_eps.endswith('.eps'):
            save_eps += '.eps'
        plt.savefig(save_eps, dpi=dpi)

    if save_svg is not None and save_svg != '':
        if not save_svg.endswith('.svg'):
            save_svg += '.svg'
        plt.savefig(save_svg, dpi=dpi)

    if save_pdf is not None and save_pdf != '':
        if not save_pdf.endswith('.pdf'):
            save_pdf += '.pdf'
        plt.savefig(save_pdf, dpi=dpi)

    if save_jpeg is not None and save_jpeg != '':
        if not save_jpeg.endswith('.jpeg'):
            save_jpeg += '.jpeg'
        plt.savefig(save_jpeg, dpi=dpi)

    if slice:
        slice_fig, slice_axes = plt.subplots(nrows=1)
        slice_plot, = slice_axes.plot([0], [0])
        mouse_event_func = lambda event: mouse_move_slice(event, slice_axes, slice_plot)
        cid = fig.canvas.mpl_connect('motion_notify_event', mouse_event_func)

    if display:
        plt.show()

    if return_plot_objects:
        return fig, axes


def varlabels_extra_axes(num_panels, this_axis, var_label):
    # apply any addition x-axes specified by the var_label keyword
    if var_label is not None:
        if not isinstance(var_label, list):
            var_label = [var_label]

        axis_delta = 0.0

        for label in var_label:
            if isinstance(label, int):
                label = tname_byindex(label)
            label_data = get_data(label, xarray=True, dt=True)

            if label_data is None:
                logging.info('Variable not found: ' + label)
                continue

            if len(label_data.values.shape) != 1:
                logging.info(
                    label + ' specified as a vector; var_label only supports scalars. Try splitting the vector into seperate tplot variables.')
                continue

            # set up the new x-axis
            axis_delta = axis_delta - num_panels * 0.1
            new_xaxis = this_axis.secondary_xaxis(axis_delta)
            xaxis_ticks = this_axis.get_xticks().tolist()
            xaxis_ticks_dt = [np.datetime64(mpl.dates.num2date(tick_val).replace(tzinfo=None).isoformat(), 'ns') for
                              tick_val in xaxis_ticks]
            # xaxis_ticks_unix = [tick_val.timestamp() for tick_val in xaxis_ticks_dt]
            xaxis_labels = get_var_label_ticks(label_data, xaxis_ticks_dt)
            new_xaxis.set_xticks(xaxis_ticks_dt)
            new_xaxis.set_xticklabels(xaxis_labels)
            ytitle = pyspedas.tplot_tools.data_quants[label].attrs['plot_options']['yaxis_opt']['axis_label']
            new_xaxis.set_xlabel(ytitle)

        # fig.subplots_adjust(bottom=0.05+len(var_label)*0.1)


def mouse_move_slice(event, slice_axes, slice_plot):
    """
    This function is called when the mouse moves over an axis
    and the slice keyword is set to True; for spectra figures, it
    updates the slice plot based on the mouse location
    """
    if event.inaxes is None:
        return

    # check for a spectrogram
    try:
        data = get_data(event.inaxes.var_name)
    except AttributeError:
        return

    if data is None:
        return

    if len(data) != 3:
        return

    slice_time = mdates.num2date(event.xdata).timestamp()
    idx = np.abs(data.times-slice_time).argmin()

    if len(data.v.shape) > 1:
        # time varying y-axis
        vdata = data.v[idx, :]
    else:
        vdata = data.v

    xaxis_options = pyspedas.tplot_tools.data_quants[event.inaxes.var_name].attrs['plot_options']['xaxis_opt']
    yaxis_options = pyspedas.tplot_tools.data_quants[event.inaxes.var_name].attrs['plot_options']['yaxis_opt']
    zaxis_options = pyspedas.tplot_tools.data_quants[event.inaxes.var_name].attrs['plot_options']['zaxis_opt']

    yrange = yaxis_options.get('y_range')
    if yrange is None:
        yrange = [np.nanmin(vdata), np.nanmax(vdata)]

    zrange = zaxis_options.get('z_range')
    if zrange is None:
        zrange = [np.nanmin(data.y), np.nanmax(data.y)]

    y_label = zaxis_options.get('axis_label')
    if y_label is not None:
        slice_axes.set_ylabel(y_label)

    title = datetime.fromtimestamp(data.times[idx], timezone.utc).strftime('%Y-%m-%d %H:%M:%S.%f')

    x_label = yaxis_options.get('axis_label')
    if x_label is not None:
        title = x_label + ' (' + title + ')'

    slice_axes.set_title(title)

    x_subtitle = yaxis_options.get('axis_subtitle')
    if x_subtitle is not None:
        slice_axes.set_xlabel(x_subtitle)

    slice_yaxis_opt = pyspedas.tplot_tools.data_quants[event.inaxes.var_name].attrs['plot_options'].get('slice_yaxis_opt')

    xscale = None
    yscale = None

    if slice_yaxis_opt is not None:
        xscale = slice_yaxis_opt.get('xi_axis_type')
        yscale = slice_yaxis_opt.get('yi_axis_type')

    if yscale is None:
        # if the user didn't explicitly set the ylog_slice option,
        # use the option from the plot
        yscale = zaxis_options.get('z_axis_type')
        if yscale is None:
            yscale = 'linear'

    if xscale is None:
        # if the user didn't explicitly set the xlog_slice option,
        # use the option from the plot
        xscale = yaxis_options.get('y_axis_type')
        if xscale is None:
            xscale = 'linear'

    if yscale == 'log' and zrange[0] == 0.0:
        zrange[0] = np.nanmin(data.y[idx, :])

    slice_plot.set_data(vdata, data.y[idx, :])
    slice_axes.set_ylim(zrange)
    slice_axes.set_xlim(yrange)
    slice_axes.set_xscale(xscale)
    slice_axes.set_yscale(yscale)

    try:
        plt.draw()
    except ValueError:
        return
    sleep(0.01)

def replace_common_exp(title):
    if hasattr(title, 'decode'):
        title = title.decode('utf-8')
    if '$' in title:
        return title
    if '^' not in title:
        return title
    exp = False
    title_out = ''
    for char in title:
        if char == '^':
            exp = True
            title_out += '$^{'
            continue
        else:
            if exp:
                if not char.isalnum():
                    title_out += '}$' + char
                    exp = False
                    continue
        title_out += char
    if exp:
        title_out += '}$'
    return title_out





import re
import logging
from pyspedas.tplot_tools import tplot_names, tplot_wildcard_expand


def tnames(pattern=None, regex=None):
    """
    Find tplot names matching a wildcard pattern.

    Parameters
    ----------
    pattern : str or list of str, optional
        Pattern(s) to search for.
        It can be a string or a list of strings (multiple patterns).
        Each pattern can contain wildcards such as * and ?, using unix-style matching.
        The default is None
    regex : str, optional
        Regular expression pattern to search for.
        If regex is provided, the pattern argument is ignored.
        The default is None.
        If both pattern and regex are None, all tplot names are returned.

    Returns
    -------
    name_list : list of str
        List of tplot variables.

    Examples
    -------
        >>> import pyspedas
        >>> import pyspedas
        >>> from pyspedas.projects.themis import fgm
        >>> fgm(trange=['2007-03-23','2007-03-24'], probe='a')
        >>> pyspedas.tnames('tha_fgs*')
        >>> pyspedas.tnames('th?_fgs_gsm')
    """
    name_list = list()
    all_names = tplot_names(quiet=True)

    if len(all_names) < 1:
        # No tplot variables found
        pass
    elif pattern is None and regex is None:
        name_list.extend(all_names)
    elif regex is not None:
        # Use re to find all names that match the regular expression
        try:
            # Check if regex is a valid regular expression
            re.compile(regex)

            for p in all_names:
                if re.match(regex, p):
                    name_list.append(p)
        except re.error:
            logging.error("Invalid regular expression.")
    elif pattern is not None:
        # Switch to tplot_wildcard_expand for improved handling of tplot names with embedded spaces
        name_list = tplot_wildcard_expand(pattern)

    return name_list



"""
This module contains the function mms_cotrans_qtransformer, which performs a quaternion rotation on a tplot variable from one coordinate system to another. It does this by recursively rotating through the ECI coordinate system.

The mms_cotrans_qtransformer function takes in four required arguments:

in_name: the name of the tplot variable to be transformed
out_name: the name of the tplot variable to be created with the transformed data
in_coord: the coordinate system of the input data
out_coord: the coordinate system to rotate to
It also takes an optional argument:

probe: the spacecraft probe number (default is '1')
The function returns the name of the output tplot variable.
"""
from pyspedas import tplot_copy, get
from .mms_cotrans_qrotate import mms_cotrans_qrotate


def mms_cotrans_qtransformer(in_name, out_name, in_coord, out_coord, probe='1'):
    """
    Support routine for mms_qcotrans; recursively goes from one coordinate system to
    another by going through ECI

    Operates on a single variable/transformation at a time

    Parameters
    -----------
        in_name: str
            Tplot variable to be transformed

        out_name: str
            Tplot variable to be created with the transformed data

        in_coord: str
            Coordinate system of the input data

        out_coord: str
            Coordinate system to rotate to

        probe: str
            Spacecraft probe #

    """

    # Final coordinate system reached
    if in_coord == out_coord:
        if in_name != out_name:
            tplot_copy(in_name, out_name)
        return out_name

    if in_coord == 'eci':
        q_name = 'mms' + probe + '_mec_quat_eci_to_' + out_coord
        q_data = get(q_name)
        if q_data is None:
            return
        mms_cotrans_qrotate(in_name, q_name, out_name, out_coord)
        recursive_in_coord = out_coord
    else:
        q_name = 'mms' + probe + '_mec_quat_eci_to_' + in_coord
        q_data = get(q_name)
        if q_data is None:
            return
        mms_cotrans_qrotate(in_name, q_name, out_name, out_coord, inverse=True)
        recursive_in_coord = 'eci'

    return mms_cotrans_qtransformer(out_name, out_name, recursive_in_coord, out_coord, probe=probe)



"""
Unit Tests for lmn transformations.
"""

import os
import logging
import unittest
import numpy as np
from pyspedas import (
    del_data,
    get_data,
    tnames,
    data_exists,
    tplot_restore,
    rotmat_get_coords,
    tplot,
)
from pyspedas.projects.omni.omni_solarwind_load import omni_solarwind_load
from pyspedas.cotrans_tools.lmn_matrix_make import lmn_matrix_make
from pyspedas.projects.themis import fgm, state
from pyspedas.utilities.config_testing import TESTING_CONFIG, test_data_download_file

# Whether to display plots during testing
global_display = TESTING_CONFIG["global_display"]
# Directory to save testing output files
output_dir = TESTING_CONFIG["local_testing_dir"]
# Ensure output directory exists
cotrans_dir = "cotrans_tools"
save_dir = os.path.join(output_dir, cotrans_dir)
if not os.path.exists(save_dir):
    os.makedirs(save_dir)
# Directory with IDL SPEDAS validation files
validation_dir = TESTING_CONFIG["remote_validation_dir"]


def _test_compare(tname1, tname2, center=False, approximate=False):
    """Compare two tplot variables."""
    try:
        data1 = get_data(tname1)
        t1 = data1.times
        y1 = data1.y
        data2 = get_data(tname2)
        t2 = data2.times
        y2 = data2.y

        if center:
            # Cut out the 10% of the edges of the data
            edge_cut = int(len(t1) * 0.1) - 1
            if edge_cut > 0 and 2 * edge_cut < len(t1):
                t1 = t1[edge_cut:-edge_cut]
                y1 = y1[edge_cut:-edge_cut]
                # find min and max of t1
                t1_min = np.min(t1)
                t1_max = np.max(t1)
                # for t2, find the indices of the values that are within t1_min and t1_max
                t2_indices = np.where((t2 >= t1_min) & (t2 <= t1_max))[0]
                t2 = t2[t2_indices]
                y2 = y2[t2_indices]

        do_match1 = np.allclose(t1, t2, rtol=1e-5, atol=1e-8)
        if approximate:
            do_match2 = np.allclose(y1, y2, rtol=1e-2, atol=1e-2)
        else:
            do_match2 = np.allclose(y1, y2, rtol=1e-5, atol=1e-8)
        do_match = do_match1 and do_match2
        if not do_match:
            logging.info(f"Data for {tname1} and {tname2} do not match.")
            logging.info(f"Data1: {data1.y}")
            logging.info(f"Data2: {data2.y}")
    except Exception as e:
        logging.info(f"Error comparing {tname1} and {tname2}: {e}")
        do_match = False
    return do_match


class TestLMNTransform(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """Find the test data files to restore."""
        # The SPEDAS script that creates the file: general/tools/python_validate/lmn_python_validate.pro

        cls.filename_sw = test_data_download_file(
            validation_dir, cotrans_dir, "solarwind_python_validate.tplot", save_dir
        )
        cls.filename_lmn = test_data_download_file(
            validation_dir, cotrans_dir, "lmn_python_validate.tplot", save_dir
        )

    # @unittest.skip("Skipping LMN tests")
    def test_lmn_1omni(self):
        """Test of OMNI variables."""
        # Load IDL variables from the test file
        del_data("*")
        tplot_restore(self.filename_sw)

        # Check if the IDL variables have been loaded
        tn = tnames()
        logging.info(tn)
        idl_bz_def = "omni_BZ_idl_default"
        idl_p_def = "omni_P_idl_default"
        idl_bz_hro = "omni_BZ_idl_2020"
        idl_p_hro = "omni_P_idl_2020"
        idl_bz_hro2 = "omni_BZ_idl_2021"
        idl_p_hro2 = "omni_P_idl_2021"
        omni_names = [
            idl_bz_def,
            idl_p_def,
            idl_bz_hro,
            idl_p_hro,
            idl_bz_hro2,
            idl_p_hro2,
        ]
        for name in omni_names:
            self.assertTrue(data_exists(name), f"Data variable {name} does not exist.")

        # Load pyspedas data (default values) and compare
        # OMNI default values (constants)
        trange = ["2050-01-01", "2050-01-01 23:59:59"]
        omni_solarwind_load(trange=trange, level="hro", suffix="_py_default")
        t1 = _test_compare(idl_bz_def, "omni_solarwind_BZ_py_default")
        self.assertTrue(t1, "Data for default BZ values do not match.")
        t2 = _test_compare(idl_p_def, "omni_solarwind_P_py_default")
        self.assertTrue(t2, "Data for default P values do not match.")

        # OMNI HRO
        trange = ["2020-01-01", "2020-01-01 23:59:59"]
        omni_solarwind_load(trange=trange, level="hro", suffix="_py_2020")
        t3 = _test_compare(idl_bz_hro, "omni_solarwind_BZ_py_2020")
        self.assertTrue(t3, "Data for 2020 HRO BZ values do not match.")
        t4 = _test_compare(idl_p_hro, "omni_solarwind_P_py_2020")
        self.assertTrue(t4, "Data for 2020 HRO P values do not match.")

        # OMNI HRO2
        trange = ["2021-01-01", "2021-01-01 06:00:00"]
        omni_solarwind_load(trange=trange, level="hro2", suffix="_py_2021")
        t5 = _test_compare(idl_bz_hro2, "omni_solarwind_BZ_py_2021")
        self.assertTrue(t5, "Data for 2021 HRO2 BZ values do not match.")
        t6 = _test_compare(idl_p_hro2, "omni_solarwind_P_py_2021")
        self.assertTrue(t6, "Data for 2021 HRO2 P values do not match.")

    # @unittest.skip("Skipping LMN tests")
    def test_lmn_2lmn(self):
        """Test of LMN transform."""
        # Load LMN data
        del_data("*")
        tplot_restore(self.filename_lmn)

        # Check if the IDL LMN variables exist in tplot
        idl_pos = "tha_state_pos_gsm_idl"
        idl_b = "tha_fgl_gsm_idl"
        idl_hro = "tha_fgl_gsm_lmn_mat_hro_idl"
        idl_p = "OMNI_solarwind_P_idl"
        idl_bz = "OMNI_solarwind_BZ_idl"
        idl_names = [idl_pos, idl_b, idl_hro, idl_p, idl_bz]
        for name in idl_names:
            self.assertTrue(data_exists(name), f"Data variable does not exist: {name}")

        # Optionally plot the IDL data
        local_png = os.path.join(save_dir, "test_lmn_1omni.png")
        tplot(idl_names, display=global_display, save_png=local_png)

        # Load pyspedas data
        trange = ["2022-01-01", "2022-01-01 06:00:00"]
        fgm(
            probe="a",
            level="l2",
            trange=trange,
            suffix="_py",
            time_clip=True,
            coord="gsm",
            get_support_data=True,
        )
        state(
            probe="a",
            trange=trange,
            suffix="_py",
            time_clip=True,
            get_support_data=True,
        )
        py_pos = "tha_pos_gsm_py"
        py_pos_i = py_pos + "_interpol"
        py_b = "tha_fgl_gsm_py"
        py_hro = "tha_fgl_gsm_lmn_mat_hro_py"
        py_p = "omni_solarwind_P"
        py_p_i = py_p + "_interpol"
        py_bz = "omni_solarwind_BZ"
        py_bz_i = py_bz + "_interpol"
        py_names = [py_pos, py_pos_i, py_b, py_hro, py_p, py_p_i, py_bz, py_bz_i]

        # Compute LMN
        lmn_matrix_make(py_pos, py_b, trange=trange, newname=py_hro)

        # Verify rotation matrix input and output coordinates
        in_coords, out_coords = rotmat_get_coords(py_hro)
        self.assertEqual(in_coords.upper(), "GSM")
        self.assertEqual(out_coords.upper(), "LMN")

        for name in py_names:
            self.assertTrue(data_exists(name), f"Data variable does not exist: {name}")
        # Optionally plot the pyspedas data
        local_png = os.path.join(save_dir, "test_lmn_2lmn.png")
        tplot(py_names, display=global_display, save_png=local_png)

        # Verify that position and B field as the same as in IDL
        t1 = _test_compare(idl_b, py_b, center=True)
        self.assertTrue(t1, "B field data does not match.")
        t2 = _test_compare(idl_pos, py_pos, center=True)
        self.assertTrue(t2, "Position data does not match.")

        # Verify that the interpolated data matches
        t3 = _test_compare(idl_p, py_p, center=True)
        self.assertTrue(t3, "OMNI P data does not match.")
        t4 = _test_compare(idl_bz, py_bz, center=True)
        self.assertTrue(t4, "OMNI Bz data does not match.")

        # Verify that the LMN matrices match
        t5 = _test_compare(idl_hro, py_hro, approximate=True)
        self.assertTrue(t5, "LMN matrices do not match for HRO data.")

        logging.info("test_lmn finished.")


if __name__ == "__main__":
    unittest.main()



import os

import mth5
import numpy as np
import pandas as pd
import pyspedas

import pyspedas
from .config import CONFIG
from pyspedas.tplot_tools import store_data, options

# MTH5 installation is checked in __init__
from mth5.clients.make_mth5 import FDSN
from mth5.mth5 import MTH5

from pyspedas.mth5.utilities import mth5_time_str
from pyspedas.mth5.utilities import _list_of_fdsn_channels

from datetime import datetime


def _disable_loguru_warnings(record):
    if record["extra"].get("no_warning"):
        return record["level"].no != mth5.logger.level("WARNING").no
    return True

def _request_df_from_input(trange, network, station):
    """
    Create a request_df from input parameters.

    Parameters
    ----------
        trange : list of str
            Time range of interest.
        network : str
            Network name.
        station : str
            Station name.

    Returns
    -------
    request_df : pandas.DataFrame
        Request dataframe.
    """

    # Get MTH5 file from the server
    # we use "*F*" channel instead of "*" to extract magnetic data
    # TODO: modify according to SEED manual appendix A: https://www.fdsn.org/pdf/SEEDManual_V2.4_Appendix-A.pdf
    channel_list = _list_of_fdsn_channels()
    # channel_list = "*F*"

    # Create request_df from input parameters
    request_df = pd.DataFrame(
        {
            "network": [network],
            "station": [station],
            "location": ["--"],
            "channel": [channel_list],
            "start": [mth5_time_str(trange[0])],  # ["2019-11-14T00:00:00"],
            "end": [mth5_time_str(trange[1])]  # ["2019-11-15T00:00:00"]
        }
    )

    return request_df


def _validate_date_format(date_string):
    """
    Validate if a date string is in the correct format.

    Parameters
    ----------
    date_string : str
        The date string to validate. Expected format: YYYY-MM-DDThh:mm:ss.

    Raises
    ------
    ValueError
        If the date string is not in the expected format.
    """
    try:
        datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%S")
    except ValueError:
        raise ValueError(f"Invalid date format: {date_string}. Expected format: YYYY-MM-DDThh:mm:ss")


def load_fdsn(trange=None, network=None, station=None,
              nodownload=False, noexception=False, print_request=False,
              nowarnings=False, request_df=None):
    """
    Load FDSN data using MTH5 interface.

    Parameters
    ----------
    trange : list of str
        Time range of interest.
    network : str
        Network name.
    station : str
        Station name.
    nodownload : bool, default=False
        If h5 file is already created do not load another one.
    noexception : bool, default=False
        If true, do not raise an exception produced by FDSN.
    print_request : bool, default=False
        Print request_df structure, which can be useful for debugging the request.
    nowarnings : bool, default=False
        If true, disable loguru warnings.
    request_df : pandas.DataFrame, optional
        Custom request_df dataframe for the `make_mth5_from_fdsn_client` method of the `FDSN` class.
        This parameter is optional, and it is not advised to use it unless the user knows exactly what to do.
        request_df must contain the following:

        - One `station`, one `network`, and one `location` as a list.
        - A list of 3 strings for `channels`.
        - The `start` and `end` date fields must contain a list of one string.

        See `pyspedas.mth5.load_fdsn._request_df_from_input` implementation for reference.

    Returns
    -------
    tplot_variable : str or None
        Tplot variable name. Tplot variable is created if data is loaded successfully, None otherwise.
    """

    # mth5.logger.remove()
    mth5.logger._core.extra["no_warning"] = nowarnings

    # If trange is not specified we don't know what to load
    if trange is None:
        pyspedas.logging.error("trange not specified")
        return

    if not network:
        pyspedas.logging.error("Network not specified")
        return

    if not station:
        pyspedas.logging.error("Station not specified")
        return

    # Handle request_df
    if request_df is None:
        request_df = _request_df_from_input(trange, network, station)
    else:
        network = request_df.network
        station = request_df.station
        trange = [request_df.start[0], request_df.end[0]]

    if print_request:
        print(request_df.to_string())

    # Create time variables that correspond to the request time period
    request_start = datetime.fromisoformat(request_df.start[0])
    request_end = datetime.fromisoformat(request_df.end[0])

    # Validate time period
    _validate_date_format(request_df.start[0])
    _validate_date_format(request_df.end[0])

    # Determine where data will be stored
    mth5dir = CONFIG['local_data_dir']
    # if os.environ.get('SPEDAS_DATA_DIR'):
    #     mth5dir = os.environ['SPEDAS_DATA_DIR']

    mth5_filename = f"{network}_{station}_{trange[0]}_{trange[1]}.h5".replace(":", "").replace("-", "")
    mth5_pathfile = os.path.join(mth5dir, mth5_filename)

    # Implementation of caching
    if nodownload and os.path.isfile(mth5_pathfile):
        mth5_path = mth5_pathfile
    else:
        # Initialize FDSN object
        fdsn_object = FDSN(mth5_version='0.2.0', client="IRIS")
        try:
            mth5_path = fdsn_object.make_mth5_from_fdsn_client(request_df, interact=False, path=mth5dir)
            if os.path.isfile(mth5_pathfile):
                pyspedas.logger.info(f"Deleting cached {mth5_pathfile}")
                os.remove(mth5_pathfile)
            pyspedas.logger.info(f"Creating cached {mth5_pathfile}")
            os.rename(mth5_path, mth5_pathfile)
            mth5_path = mth5_pathfile
        except Exception as e:
            # Hande mth5 object initialization error.
            pyspedas.logger.error(f"Cannot initialize mth5 object:\n{e}")

            # # Check if file was the cache file was created
            # if 'mth5_path' in locals() and os.path.isfile(mth5_path):
            #     pyspedas.logger.info(f"Cache was created and will be deleted: {mth5_path}")
            #     os.remove(mth5_path)

            # Exit code by flag
            if noexception:
                return
            raise
        finally:
            mth5_tmp = fdsn_object.make_filename(request_df)
            mth5_file = os.path.join(mth5dir, mth5_tmp)
            if os.path.isfile(mth5_file):
                pyspedas.logger.info(f"Deleting mth5 temporary h5 file {mth5_tmp}")
                os.remove(mth5_file)

    # Using MTH5 as a context manager
    with MTH5() as mth5_object:
        mth5_object.open_mth5(mth5_path)
        surveys = mth5_object.surveys_group.groups_list

        # mth5_object = MTH5()
        # mth5_object.open_mth5(mth5_path)
        # surveys = mth5_object.surveys_group.groups_list
        # # print(surveys)

        # Create new data for time, x, y and z
        time = np.array([])
        attributes = [('hx', 'bx'), ('hy', 'by'), ('hz', 'bz')]
        variables = {'x': np.array([]), 'y': np.array([]), 'z': np.array([])}

        attr_dict = {'filename': mth5_path}  # metadata with filename
        units = set()  # basic units
        measurements_type = set()  # basic type of measurements

        # loop over surveys and runs
        for survey in surveys:
            station_data = mth5_object.get_station(station_name=station, survey=survey)
            attr_dict["metadata"] = station_data.metadata

            # TODO: Check orientation reference frame
            # "orientation.reference_frame": "geographic"

            # Loop over runs
            for run_name in station_data.metadata.run_list:
                # Get the data
                run_data = station_data.get_run(run_name=run_name)

                # Check run time coverage
                run_start = datetime.fromisoformat(run_data.metadata.time_period.start).replace(tzinfo=None)
                run_end = datetime.fromisoformat(run_data.metadata.time_period.end).replace(tzinfo=None)

                # Skip processing if run is outside requested time range
                if run_start > request_end or run_end < request_start or request_end <= request_start:
                    continue

                # Get the run table summary
                # Time clips according to original times
                run_ts = run_data.to_runts(start=request_df.start[0], end=request_df.end[0])

                # Determine is we have all components
                # can channels_recorded_magnetic be missing from the dataset?
                # TODO: What to do with the runs without 3 components - operate only with 3 components for now...
                if 'channels_recorded_magnetic' in run_ts.dataset.attrs:
                    if len(run_ts.dataset.channels_recorded_magnetic) != 3:
                        pyspedas.logger.info("Run has less than 3 components")
                        continue
                else:
                    pyspedas.logger.warning("Run dataset has no channels_recorded_magnetic attribute")
                    continue

                # Check if the filters are available to calibrate data
                if len(run_ts.filters) == 0:
                    pyspedas.logger.warning("Filters were not added to run_ts")
                    continue

                # Convert counts to physical units
                run_ts = run_ts.calibrate()

                # Copy numpy arrays to ensure that data is in memory after mth5 is closed
                if time.size == 0:
                    time = run_ts.dataset.time.to_numpy()
                else:
                    time = np.append(time, run_ts.dataset.time.to_numpy())

                # Check if all required attributes are present
                all_attributes_present = all(
                    hasattr(run_ts.dataset, attr_h) or hasattr(run_ts.dataset, attr_b)
                    for attr_h, attr_b in attributes
                )
                if all_attributes_present:
                    for (attr_h, attr_b), var_name in zip(attributes, variables.keys()):
                        if hasattr(run_ts.dataset, attr_h):
                            data = getattr(run_ts.dataset, attr_h)
                        elif hasattr(run_ts.dataset, attr_b):
                            data = getattr(run_ts.dataset, attr_b)
                        else:
                            pyspedas.logger.warning(f"Neither {attr_h} nor {attr_b} found in run dataset")
                            continue

                        variables[var_name] = np.append(variables[var_name], data.to_numpy())

                        try:
                            units.add(data.units)
                            measurements_type.add(data.type)
                        except AttributeError:
                            pyspedas.logger.warning(f"Problem with adding {attr_h}/{attr_b} dataset units or type")

                    # After the loop, you can reassign the modified arrays back to x, y, z
                    # x, y, z = variables['x'], variables['y'], variables['z']
                else:
                    pyspedas.logger.warning("Some of the attributes are not present in run_ts dataset")
                    continue

        # TODO: Should data be sorted before saving
        data = {'x': time, 'y': np.vstack((variables['x'], variables['y'], variables['z'])).T}
        tplot_variable = 'fdsn_' + network + '_' + station

        # Handle legends
        legend_names = ['x', 'y', 'z']
        try:
            legend_names = station_data.metadata.channels_recorded
        except AttributeError or UnboundLocalError:
            pyspedas.logger.error("station_data was not defined correctly")
            # Also raise this exception if noexception flag is set
            if not noexception:
                raise

        tplot_variable_created = False
        try:
            store_data(tplot_variable, data=data, attr_dict=attr_dict)
            tplot_variable_created = True
        except IndexError:
            pyspedas.logger.error(f"Cannot create empty tplot variabe {tplot_variable}")
        except Exception as E:
            pyspedas.logger.error(f"Unexpected error {E} while creating tplot variable {tplot_variable}")
            if not noexception:
                raise

        if tplot_variable_created:
            tplot_options = {
                "name": f"FDSN: {network}, station: {station}",  # network and stations
                "ytitle": f"{','.join(list(measurements_type))}",  # measurement
                "ysubtitle": f"{','.join(list(units))}",  # units
                "legend_names": legend_names
            }

            options(tplot_variable, opt_dict=tplot_options)

            # TODO: For potential future implementation
            #     filename : bool, default=False
            #         If true, return filename as a second return parameter

            return tplot_variable



# Copyright 2020 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import logging
import pyspedas
from pyspedas.tplot_tools import store_data
from copy import deepcopy
from collections import OrderedDict


def tplot_copy(old_name, new_name):
    """
    This function will copy a tplot variables that is already stored in memory.

    Parameters
    ----------
        old_name : str
            Old name of the Tplot Variable
        new_name : str
            Name of the copied Tplot Variable

    Returns
    -------
        None

    Examples
    --------
        >>> # Copy Variable 1 into a new Variable 2
        >>> import pyspedas
        >>> pyspedas.tplot_copy("Variable1", "Variable2")

    """

    # if old name input is a number, convert to corresponding name
    if isinstance(old_name, int):
        if isinstance(pyspedas.tplot_tools.data_quants[old_name], dict):
            old_name = pyspedas.tplot_tools.data_quants[old_name]['name']
        else:
            old_name = pyspedas.tplot_tools.data_quants[old_name].name

    # check if old name is in current dictionary
    if old_name not in pyspedas.tplot_tools.data_quants.keys():
        logging.info("The name %s is currently not in pyspedas",old_name)
        return

    # Add a new data quantity with the copied data
    if isinstance(pyspedas.tplot_tools.data_quants[old_name], dict):
        # old variable is a non-record varying variable
        store_data(new_name, data={'y': pyspedas.tplot_tools.data_quants[old_name]['data']})
    else:
        pyspedas.tplot_tools.data_quants[new_name] = deepcopy(pyspedas.tplot_tools.data_quants[old_name])
        pyspedas.tplot_tools.data_quants[new_name].name = new_name

    return



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import datetime
from dateutil.parser import parse

def str_to_float_fuzzy(time_str):
    """
    Implementation of str_to_int (below) that uses dateutil and .timestamp()
    to convert the date/time string to integer (number of seconds since Jan 1, 1970)

    This function is slower than str_to_int, but more flexible
    """
    dt_object = parse(time_str)
    return dt_object.replace(tzinfo=datetime.timezone.utc).timestamp()



from .load import load
from pyspedas.utilities.datasets import find_datasets

# This routine was moved out of __init__.py. Please see that file for previous revision history.

def datasets(instrument=None, label=True):
    return find_datasets(mission='Wind', instrument=instrument, label=label)



import unittest
import os
from pyspedas.projects.mms.mms_orbit_plot import mms_orbit_plot


class TestMMSOrbitPlot(unittest.TestCase):
    def test_trange(self):
        # Set trange and save_png options
        trange = ['2015-10-16/00:00:00', '2015-10-16/12:00:00']
        save_png = 'test_trange'

        mms_orbit_plot(trange=trange, save_png=save_png, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_png + '.png'))

    def test_probes(self):
        # Set probes and save_png options
        probes = [1, 2]
        save_png = 'test_probes'

        mms_orbit_plot(probes=probes, save_png=save_png, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_png + '.png'))

    def test_data_rate(self):
        # Set data_rate and save_png options
        data_rate = 'brst'
        save_png = 'test_data_rate'

        mms_orbit_plot(probes=[1, 4], trange=['2019-05-01/02:00', '2019-05-01/02:20'], data_rate=data_rate, save_png=save_png, display=False, earth=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_png + '.png'))

    def test_plane(self):
        # Set plane and save_png options
        plane = 'yz'
        save_png = 'test_plane'

        mms_orbit_plot(plane=plane, save_png=save_png, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_png + '.png'))

    def test_xr(self):
        # Set xr and save_png options
        xr = [-5, 5]
        save_png = 'test_xr'

        mms_orbit_plot(xr=xr, save_png=save_png, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_png + '.png'))

    def test_yr(self):
        # Set yr and save_png options
        yr = [-5, 5]
        save_png = 'test_yr'

        mms_orbit_plot(yr=yr, save_png=save_png, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_png + '.png'))

    def test_coord(self):
        # Set coord and save_png options
        coord = 'gsm'
        save_png = 'test_coord'

        mms_orbit_plot(coord=coord, save_png=save_png, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_png + '.png'))

    def test_xsize(self):
        # Set xsize and save_png options
        xsize = 10
        save_png = 'test_xsize'

        mms_orbit_plot(xsize=xsize, save_png=save_png, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_png + '.png'))

    def test_ysize(self):
        # Set ysize and save_png options
        ysize = 10
        save_png = 'test_ysize'

        mms_orbit_plot(ysize=ysize, save_png=save_png, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_png + '.png'))

    def test_marker(self):
        # Set marker and save_png options
        marker = 'o'
        save_png = 'test_marker'

        mms_orbit_plot(marker=marker, save_png=save_png, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_png + '.png'))

    def test_markevery(self):
        # Set markevery and save_png options
        markevery = 5
        save_png = 'test_markevery'

        mms_orbit_plot(markevery=markevery, save_png=save_png, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_png + '.png'))

    def test_markersize(self):
        # Set markersize and save_png options
        markersize = 10
        save_png = 'test_markersize'

        mms_orbit_plot(markersize=markersize, save_png=save_png, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_png + '.png'))

    def test_earth(self):
        # Set earth and save_png options
        earth = False
        save_png = 'test_earth'

        mms_orbit_plot(earth=earth, save_png=save_png, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_png + '.png'))

    def test_dpi(self):
        # Set dpi and save_png options
        dpi = 100
        save_png = 'test_dpi'

        mms_orbit_plot(dpi=dpi, save_png=save_png, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_png + '.png'))

    def test_save_pdf(self):
        # Set save_png option
        save_pdf = 'test_save_pdf'

        mms_orbit_plot(save_pdf=save_pdf, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_pdf + '.pdf'))

    def test_save_eps(self):
        # Set save_eps option
        save_eps = 'test_save_eps'

        mms_orbit_plot(save_eps=save_eps, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_eps + '.eps'))

    def test_save_jpeg(self):
        # Set save_jpeg option
        save_jpeg = 'test_save_jpeg'

        mms_orbit_plot(save_jpeg=save_jpeg, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_jpeg + '.jpeg'))

    def test_save_svg(self):
        # Set save_svg option
        save_svg = 'test_save_svg'

        mms_orbit_plot(save_svg=save_svg, display=False)

        # Check that the test figure was saved
        self.assertTrue(os.path.exists(save_svg + '.svg'))

    def test_return_plot_objects(self):
        # Set return_plot_objects option
        return_plot_objects = True

        plot_objects = mms_orbit_plot(return_plot_objects=return_plot_objects, display=False)

        # Check that plot_objects is a tuple
        self.assertIsInstance(plot_objects, tuple)


if __name__ == '__main__':
    unittest.main()



"""
This is a python version of the wavelet.pro script in IDL SPEDAS.
It computes the wavelet transform as described by Torrence & Compo (1998).

Reference: Torrence, C. and G. P. Compo, 1998: A Practical Guide to
           Wavelet Analysis. <I>Bull. Amer. Meteor. Soc.</I>, 79, 61-78.
"""

import numpy as np
from scipy.special import gamma, factorial
from scipy.stats import chi2


# Morlet wavelet function (Fourier domain)
def morlet(k0, scale, k):
    """
    Morlet wavelet in the Fourier domain.

    Parameters
    ----------
    k0 : float
        Non-dimensional wavenumber of the Morlet wavelet. If set to -1,
        the function will use the default value of 6.0.
    scale : float
        Wavelet scale.
    k : array_like
        Array of angular wavenumbers (rad / time unit). Typically created
        as in wavelet98: k = [0, k_pos..., k_neg...].

    Returns
    -------
    vmorlet : ndarray
        Complex-valued Morlet wavelet evaluated in the Fourier domain at
        the frequencies given by `k`. Shape matches `k`.
    period : float
        Equivalent Fourier period corresponding to the provided `scale`.
    coi : float
        Cone-of-influence value for this scale (in same units as `period`).
    dofmin : int
        Minimum degrees of freedom for significance testing (no smoothing).
    cdelta : float
        Reconstruction factor used for inverse wavelet reconstruction.
        May be -1 if undefined for the chosen `k0`.
    psi0 : float
        Time-domain normalization constant of the mother wavelet.

    Notes
    -----
    Implementation follows Torrence & Compo (1998). The returned `vmorlet`
    is normalized such that total energy equals the signal length N (see
    inline code comment "total energy=N").
    """
    # Set default wavenumber if not provided
    if k0 == -1:
        k0 = 6.0
    n = len(k)
    # Exponent for Morlet wavelet in Fourier space
    expnt = -((scale * k - k0) ** 2) / 2.0 * (k > 0)
    # Time step in Fourier domain
    dt = 2 * np.pi / (n * k[1])
    # Normalization factor
    norm = np.sqrt(2 * np.pi * scale / dt) * (
        np.pi**-0.25
    )  # total energy=N   [Eqn (7)]
    # Morlet wavelet in Fourier space
    vmorlet = norm * np.exp(np.maximum(expnt, -100))
    vmorlet = vmorlet * (expnt > -100)  # avoid underflow errors
    vmorlet = vmorlet * (k > 0)  # Heaviside step function
    # Scale to period conversion factor
    fourier_factor = (4 * np.pi) / (k0 + np.sqrt(2 + k0**2))  # Scale-->Fourier [Sec.3h]
    period = scale * fourier_factor
    # Cone of influence
    coi = fourier_factor / np.sqrt(2)  # Cone-of-influence [Sec. 3g]
    dofmin = 2  # Degrees of freedom

    cdelta = 0.776 if k0 == 6 else -1  # Reconstruction factor
    psi0 = np.pi**-0.25

    return vmorlet, period, coi, dofmin, cdelta, psi0


# Paul wavelet function (Fourier domain)
def paul(m, scale, k):
    """
    Paul wavelet in the Fourier domain.

    Parameters
    ----------
    m : float
        Order of the Paul wavelet. If set to -1, the function will use the
        default value of 4.0.
    scale : float
        Wavelet scale.
    k : array_like
        Array of angular wavenumbers (rad / time unit). Typically created
        as in wavelet98: k = [0, k_pos..., k_neg...].

    Returns
    -------
    vpaul : ndarray
        Complex-valued Paul wavelet evaluated in the Fourier domain at
        the frequencies given by `k`. Shape matches `k`.
    period : float
        Equivalent Fourier period corresponding to the provided `scale`.
    coi : float
        Cone-of-influence value for this scale (in same units as `period`).
    dofmin : int
        Minimum degrees of freedom for significance testing (no smoothing).
    cdelta : float
        Reconstruction factor used for inverse wavelet reconstruction.
        May be -1 if undefined for the chosen `m`.
    psi0 : float
        Time-domain normalization constant of the mother wavelet.

    Notes
    -----
    Implementation follows Torrence & Compo (1998). The returned `vpaul`
    is normalized such that total energy equals the signal length N (see
    inline code comment "total energy=N").
    """
    # Set default order if not provided
    if m == -1:
        m = 4.0
    n = len(k)
    expnt = -(scale * k) * (k > 0)
    dt = 2 * np.pi / (n * k[1])
    norm = np.sqrt(2 * np.pi * scale / dt) * (2**m / np.sqrt(m * factorial(2 * m - 1)))
    vpaul = norm * ((scale * k) ** m) * np.exp(np.maximum(expnt, -100)) * (expnt > -100)
    vpaul = vpaul * (k > 0)
    fourier_factor = 4 * np.pi / (2 * m + 1)
    period = scale * fourier_factor
    coi = fourier_factor * np.sqrt(2)
    dofmin = 2  # Degrees of freedom with no smoothing
    cdelta = 1.132 if m == 4 else -1  # Reconstruction factor
    psi0 = 2**m * factorial(m) / np.sqrt(np.pi * factorial(2 * m))

    return vpaul, period, coi, dofmin, cdelta, psi0


# Derivative of Gaussian (DOG) wavelet function (Fourier domain)
def dog(m, scale, k):
    """
    Derivative of Gaussian (DOG) wavelet in the Fourier domain.

    Parameters
    ----------
    m : int or float
        Order of the DOG wavelet (m-th derivative). If set to -1, the
        function will use the default value of 2.
    scale : float
        Wavelet scale.
    k : array_like
        Array of angular wavenumbers (rad / time unit). Typically created
        as in this module: k = [0, k_pos..., k_neg...].

    Returns
    -------
    gauss : ndarray
        Complex-valued DOG wavelet evaluated in the Fourier domain at the
        frequencies given by `k`. Shape matches `k`.
    period : float
        Equivalent Fourier period corresponding to the provided `scale`.
    coi : float
        Cone-of-influence value for this scale (in same units as `period`).
    dofmin : int
        Minimum degrees of freedom for significance testing (no smoothing).
    cdelta : float
        Reconstruction factor used for inverse wavelet reconstruction.
        May be -1 if undefined for the chosen `m`.
    psi0 : float
        Time-domain normalization constant of the mother wavelet.

    Notes
    -----
    Implementation follows Torrence & Compo (1998). Returned `gauss` is
    normalized such that total energy equals the signal length when used
    consistently with the rest of the transform implementation in this file.
    """
    # Set default order if not provided
    if m == -1:
        m = 2
    n = len(k)
    expnt = -((scale * k) ** 2) / 2.0
    dt = 2 * np.pi / (n * k[1])
    norm = np.sqrt(2 * np.pi * scale / dt) * np.sqrt(1.0 / gamma(m + 0.5))
    gauss = (
        -norm
        * (1j**m)
        * (scale * k) ** m
        * np.exp(np.maximum(expnt, -100))
        * (expnt > -100)
    )
    fourier_factor = 2 * np.pi * np.sqrt(2.0 / (2 * m + 1))
    period = scale * fourier_factor
    coi = fourier_factor / np.sqrt(2)
    dofmin = 1  # Degrees of freedom with no smoothing
    cdelta = -1
    psi0 = -1
    if m == 2:
        cdelta = 3.541  # reconstruction factor
        psi0 = 0.867325
    if m == 6:
        cdelta = 1.966  # reconstruction factor
        psi0 = 0.88406

    return gauss, period, coi, dofmin, cdelta, psi0


# Main wavelet transform function (Torrence & Compo 1998 style)
def wavelet98(
    y1,
    dt,
    s0=None,
    dj=None,
    j_scales=None,
    pad=0,
    mother="MORLET",
    param=-1,
    lag1=0.0,
    siglvl=0.95,
    recon=False,
    fft_theor=None,
    do_wave=True,
    do_daughter=False,
):
    """
    Compute the wavelet transform of a 1D time series
    using the Torrence & Compo (1998) algorithm.

    Parameters
    ----------
    y1 : array_like
        Time series to be analyzed.
    dt : float
        Time step of the time series (sampling time).
    s0 : float, optional
        Smallest scale of the wavelet transform. Default is 2 * dt.
    dj : float, optional
        Spacing between discrete scales of the wavelet transform. Default is 0.125.
        A smaller value will give better scale resolution, but it will be slower.
    j_scales : int, optional
        Number of scales minus one to use in the wavelet transform.
        Scales range from s0 to s0 * 2^(j_scales * dj).
        Default is int(np.floor(np.log2(n * dt / s0) / dj)).
    pad : int {0, 1, 2}, optional::

            pad = 0, no padding (default)
            pad = 1, pad the time series to the next power of 2
            pad = 2, pad the time series to the next power of 4

    mother : string {'MORLET', 'PAUL', 'DOG'}, optional
        A string giving the mother wavelet to use. Default is 'MORLET'.
    param : float, optional
        Parameter for the mother wavelet. Default is -1::

                For 'Morlet' this is k0 (wavenumber), default is 6.
                For 'Paul' this is m (order), default is 4.
                For 'DOG' this is m (m-th derivative), default is 2.

    lag1 : float, optional
        Lag-1 autocorrelation of the time series used for SIGNIF levels.
        Default is 0.0.
    siglvl : float, optional
        Significance level to use for the wavelet transform. Default is 0.95.
    recon : bool, optional
        If True, reconstruct the time series from the wavelet transform.
        Default is False.
    fft_theor : array_like, optional
        Theoretical background spectrum as a function of
        Fourier frequency. This will be smoothed by the
        wavelet function.
    do_wave: bool, optional
        If True, return the wavelet transform. Default is True.
    do_daughter: bool, optional
        If True, return the wavelet transform of the mother wavelet.
        Default is False.

    Returns
    -------
    wave : array_like
        Wavelet transform of the time series.
        This is a complex array of dimensions (N,j_scales+1).
    scale : array_like
        Scales of the wavelet transform.
    period : array_like
        Periods of the wavelet transform.
    signif : array_like
        Significance levels of the wavelet transform.
    daughter: array_like
        Wavelet wavelets, if do_daughter is True. Else, an empty array.
    y2 : array_like
        Reconstructed time series, if recon is True. Else, -1.
    fft_theor_out : array_like
        Output theoretical background spectrum (smoothed by the
        wavelet function) that was used in the wavelet transform.

    References
    ----------
    Torrence, C. and G. P. Compo, 1998: A Practical Guide to Wavelet Analysis.
    <I>Bull. Amer. Meteor. Soc.</I>, 79, 61-78.
    """
    y1 = np.asarray(y1)
    n = len(y1)
    n1 = n
    y2 = -1  # This will become the reconstructed time series, if recon is True
    y1_orig = y1.copy()

    # Set default parameters if not provided
    if s0 is None:
        s0 = 2 * dt
    if dj is None:
        dj = 1.0 / 8
    if j_scales is None:
        j_scales = int(np.trunc(np.log2(n * dt / s0) / dj))  # [Eqn (10)]
    if mother is None:
        mother = "MORLET"
    if param is None:
        param = -1

    # Remove mean from input
    ypad = y1 - np.sum(y1) / n

    # Pad to next power of 2 for FFT speed, if requested
    if pad is not None and pad != 0:
        n = len(ypad)
        if pad == 1:
            base2 = int(np.trunc(np.log2(n) + 0.4999))  # for "nearest power of 2"
        elif pad == 2:
            base2 = int(np.trunc(np.log2(n)))  # for "next lowest integer power of 2"
        else:
            raise ValueError("pad must be 1 or 2")
        num_pad = 2 ** (base2 + 1) - n

        # Pad the input
        if num_pad > 0:
            ypad = np.concatenate([ypad, np.zeros(num_pad, dtype=np.float32)])
        n = len(ypad)

    # Construct scale array (logarithmic spacing)
    na = int(j_scales + 1)  # Cast na as an integer, this is the number of scales
    scale = np.arange(na) * dj  # array of j values
    scale = (2.0**scale) * s0  # array of scales 2^j, [Eqn (9)]
    period = np.empty(na, dtype=np.float32)  # uninitialized (empty) float array
    wave = np.empty((n, na), dtype=np.complex64)  # uninitialized (empty) complex array
    daughter = np.empty(
        (n, na), dtype=np.complex64
    )  # uninitialized (empty) complex array

    # Construct wavenumber array used in transform [Eqn (5)]
    k_pos = (np.arange(n // 2, dtype=np.float64) + 1) * (2 * np.pi) / (n * dt)
    k_reversed = -k_pos[: ((n - 1) // 2)][::-1]
    k = np.concatenate(([0.0], k_pos, k_reversed))

    # FFT of the (possibly padded) input, normalize it similarly to IDL
    yfft = np.fft.fft(ypad, norm="forward")  # [Eqn (3)]

    # Theoretical background spectrum for significance
    if fft_theor is not None and len(fft_theor) == n:
        # If fft_theor exists and has length n, use it
        fft_theor_k = fft_theor
    else:
        fft_theor_k = (1 - lag1**2) / (
            1 - 2 * lag1 * np.cos(k * dt) + lag1**2
        )  # [Eqn(16)]
    fft_theor_out = np.zeros(na)

    # Main loop over all scales
    for a1 in range(na):
        if mother.upper() == "MORLET":
            psi_fft, period1, coi, dofmin, cdelta, psi0 = morlet(param, scale[a1], k)
        elif mother.upper() == "PAUL":
            psi_fft, period1, coi, dofmin, cdelta, psi0 = paul(param, scale[a1], k)
        elif mother.upper() == "DOG":
            psi_fft, period1, coi, dofmin, cdelta, psi0 = dog(param, scale[a1], k)
        else:
            raise ValueError(
                "Unknown mother wavelet; accepted values: MORLET, PAUL, DOG"
            )
        # Inverse FFT to get wavelet transform at this scale
        if do_wave:
            wave[:, a1] = np.fft.ifft(yfft * psi_fft) * n  # wavelet transform [Eqn (4)]
        if do_daughter:
            daughter[:, a1] = np.fft.ifft(psi_fft) * n

        period[a1] = period1
        # Theoretical spectrum for significance
        fft_theor_out[a1] = np.sum(np.abs(psi_fft) ** 2 * fft_theor_k) / n

    #  ; COI [Sec.3g]
    idx = np.concatenate(
        [
            np.arange((n1 + 1) // 2, dtype=np.float32),  # FINDGEN((n1+1)/2)
            np.arange(n1 // 2 - 1, -1, -1, dtype=np.float32),  # REVERSE(FINDGEN(n1/2))
        ]
    )
    # Elementwise multiply and scale by dt
    coi = coi * idx * dt

    # shift so DAUGHTERs are in middle of array
    if do_daughter:
        half_n1 = n1 // 2
        daughter = np.concatenate(
            (daughter[n - half_n1 :, ...], daughter[: n - half_n1, ...]), axis=0
        )

    # Significance levels (chi-squared test) [Sec.4]
    sdev = np.var(y1_orig)
    fft_theor_out = sdev * fft_theor_out
    dof = dofmin
    # IDL uses the chisqr_csf function, the equivalent in python is scipy.stats.chi2.ppf
    signif = fft_theor_out * chi2.isf(1.0 - siglvl, dof) / dof  # [Eqn (18)]

    # Optional: reconstruct the time series from the wavelet transform [Eqn (11)]
    if recon:
        if cdelta == -1:
            y2 = -1
            print("cdelta undefined, cannot reconstruct with this wavelet")
        else:
            wave_float = (
                wave.astype(np.float32)
                if not np.issubdtype(wave.dtype, np.floating)
                else wave
            )

            # Refine reconstruction logic to match IDL
            scale_factor = 1.0 / np.sqrt(scale)
            y2 = (
                dj
                * np.sqrt(dt)
                / (cdelta * psi0)
                * np.sum(wave_float * scale_factor[None, :], axis=1)
            )
            y2 = y2[:n1]  # Ensure the reconstructed array matches the original length

    # Return wavelet transform, scales, periods, and significance
    return wave[:n1, :], scale, period, signif, daughter, y2, fft_theor_out



import numpy as np

from copy import deepcopy

from pyspedas import tinterpol
from pyspedas.tplot_tools import time_double
from pyspedas.tplot_tools import time_string
from pyspedas import tnames

from pyspedas.particles.moments.spd_pgs_moments import spd_pgs_moments
from pyspedas.particles.spd_part_products.spd_pgs_regrid import spd_pgs_regrid
from pyspedas import get_timespan, get_data, store_data

from .erg_xep_get_dist import erg_xep_get_dist
from .erg_pgs_clean_data import erg_pgs_clean_data
from .erg_pgs_limit_range import erg_pgs_limit_range
from .erg_convert_flux_units import erg_convert_flux_units
from .erg_pgs_moments_tplot import erg_pgs_moments_tplot
from .erg_pgs_make_fac import erg_pgs_make_fac
from .erg_pgs_make_e_spec import erg_pgs_make_e_spec
from .erg_pgs_make_theta_spec import erg_pgs_make_theta_spec
from .erg_pgs_make_phi_spec import erg_pgs_make_phi_spec
from .erg_pgs_do_fac import erg_pgs_do_fac
from .erg_pgs_progress_update import erg_pgs_progress_update
from .erg_pgs_make_tplot import erg_pgs_make_tplot

def erg_xep_part_products(
    in_tvarname,
    species=None,
    outputs=['energy'],
    no_ang_weighting=True,
    suffix='',
    units='flux',
    datagap=15.0,
    regrid=[16, 12],
    pitch=[0., 180.],
    theta=[-90., 90.],
    phi_in=[0., 360.],
    gyro=[0., 360.],
    energy=None,
    fac_type='mphism',
    trange=None,
    mag_name=None,
    pos_name=None,
    relativistic=False,
    no_regrid=False
    ):

    if len(tnames(in_tvarname)) < 1:
        print('No input data, please specify tplot variable!')
        return 0

    in_tvarname = tnames(in_tvarname)[0]
    instnm = in_tvarname.split('_')[1]  #  ;; xep

    
    #  ;; no_regrid is always on, otherwise QHULL hangs in gridding
    no_regrid = True
    
    if no_ang_weighting:
        no_regrid = True

    if isinstance(outputs, str):
        outputs_lc = outputs.lower()
        outputs_lc = outputs_lc.split(' ')
    elif isinstance(outputs, list):
        outputs_lc = []
        for output in outputs:
            outputs_lc.append(output.lower())

    units_lc = units.lower()

    if phi_in != [0., 360.]:
        if abs(phi_in[1] - phi_in[0]) > 360.:
            print('ERROR: Phi restrictions must have range no larger than 360 deg')
            return 0
        
        phi = phi_in  # Survey or implement of spd_pgs_map_azimuth() have not conducted yet.
        if phi[0] == phi[1]:
            phi = [0., 360.]


    if abs(gyro[1] - gyro[0]) > 360.:
        print('ERROR: Gyro restrictions must have range no larger than 360 deg')
        return 0
    if gyro[0] == gyro[1]:
        gyro = [0., 360.]


    """
    ;;Create energy spectrogram after FAC transformation if limits are not 
    ;;identical to the default.
    """
    if (gyro != [0., 360.]) or (pitch != [0.,180.]):
        idx = np.where(np.array(outputs_lc) == 'energy')[0]
        if (idx.shape[0] > 0) and ('fac_energy' not in outputs_lc):
            idx = idx[0]
            outputs_lc[idx] = 'fac_energy'

        idx = np.where(np.array(outputs_lc) == 'moments')[0]
        if (idx.shape[0] > 0) and ('fac_moments' not in outputs_lc):
            idx = idx[0]
            outputs_lc[idx] = 'fac_moments'

    #  ;;Preserve the original time range
    tr_org = get_timespan(in_tvarname)

    if instnm == 'xep':
        times_array = erg_xep_get_dist(in_tvarname, species=species, units=units_lc, time_only=True)
    else:
        print(f'ERROR: Cannot find "xep" in the given tplot variable name: {in_tvarname}')
        return 0

    if trange is not None:
        
        trange_double = time_double(trange)
        time_indices = np.where((times_array >= trange_double[0]) \
                                & (times_array <= trange_double[1]))[0]
        
        if time_indices.shape[0] < 1:
            print(f'No ,{in_tvarname}, data for time range ,{time_string(trange_double)}')

    elif trange is None:
        time_indices = np.arange(times_array.shape[0])

    times_array = times_array[time_indices]



    if instnm == 'xep':
        dist = erg_xep_get_dist(in_tvarname, 0, species=species, units=units_lc)

    if 'energy' in outputs_lc:
        out_energy = np.zeros((times_array.shape[0], dist['n_energy']))
        out_energy_y = np.zeros((times_array.shape[0], dist['n_energy']))
    if 'phi' in outputs_lc:
        out_phi = np.zeros((times_array.shape[0], dist['n_phi']))
        out_phi_y = np.zeros((times_array.shape[0], dist['n_phi']))
    if 'pa' in outputs_lc:
        out_pad = np.zeros((times_array.shape[0], regrid[1]))
        out_pad_y = np.zeros((times_array.shape[0], regrid[1]))

    if 'fac_energy' in outputs_lc:
        out_fac_energy = np.zeros((times_array.shape[0], dist['n_energy']))
        out_fac_energy_y = np.zeros((times_array.shape[0], dist['n_energy']))

    out_vars = []
    last_update_time = None

    """
    ;;--------------------------------------------------------
    ;;Prepare support data
    ;;--------------------------------------------------------
    """
    # ;;create rotation matrix to B-field aligned coordinates if needed
    
    fac_outputs = ['pa', 'fac_energy']
    fac_requested = len(set(outputs_lc).intersection(fac_outputs)) > 0
    if fac_requested:

        fac_matrix = erg_pgs_make_fac(times_array, mag_name, pos_name, fac_type=fac_type)

        if fac_matrix is None:
            # problem creating the FAC matrices
            fac_requested = False

    magf = np.array([0., 0., 0.])
    no_mag_for_moments = False

    """
    ;;-------------------------------------------------
    ;; Loop over time to build spectrograms and/or moments
    ;;-------------------------------------------------
    """
    for index in range(time_indices.shape[0]):

        last_update_time = erg_pgs_progress_update(last_update_time=last_update_time,
             current_sample=index, total_samples=time_indices.shape[0], type_string=in_tvarname)

        #  ;; Get the data structure for this sample

        if instnm == 'xep':
            dist = erg_xep_get_dist(in_tvarname, time_indices[index], species=species, units=units_lc)
        else:
            print(f'ERROR: Cannot find "xep" in the given tplot variable name: {in_tvarname}')
            return 0
        if magf.ndim == 2:
            magvec = magf[index]
        elif magf.ndim == 1:
            magvec = magf

        clean_data = erg_pgs_clean_data(dist, units=units_lc,relativistic=relativistic, magf=magvec)

        if fac_requested:
            pre_limit_bins = deepcopy(clean_data['bins'])

        clean_data = erg_pgs_limit_range(clean_data, phi=phi_in, theta=theta, energy=energy, no_ang_weighting=no_ang_weighting)

        #  ;;Build energy spectrogram
        if 'energy' in outputs_lc:
            out_energy_y[index, :], out_energy[index, :] = erg_pgs_make_e_spec(clean_data)

        #  ;;Build phi spectrogram
        if 'phi' in outputs_lc:
            out_phi_y[index, :], out_phi[index, :] = erg_pgs_make_phi_spec(clean_data, resolution=dist['n_phi'],no_ang_weighting=no_ang_weighting)

        #  ;;Perform transformation to FAC, (regrid data), and apply limits in new coords
        
        if fac_requested:
            
            # ;limits will be applied to energy-aligned bins
            clean_data['bins'] = deepcopy(pre_limit_bins)
            clean_data = erg_pgs_limit_range(clean_data, phi=phi_in, theta=theta, energy=energy, no_ang_weighting=no_ang_weighting)

            # ;perform FAC transformation and interpolate onto a new, regular grid 
            clean_data = erg_pgs_do_fac(clean_data, fac_matrix[index, :, :])

            #;nearest neighbor interpolation to regular grid in FAC
            if not no_regrid:
                if (not np.all(np.isnan(clean_data['theta']))) and (not np.all(np.isnan(clean_data['phi']))):
                    clean_data = spd_pgs_regrid(clean_data, regrid)

            clean_data['theta'] = 90.0-clean_data['theta']  #  ;pitch angle is specified in co-latitude

            # ;apply gyro & pitch angle limits(identical to phi & theta, just in new coords)
            clean_data = erg_pgs_limit_range(clean_data, theta=pitch, phi=gyro, no_ang_weighting=no_ang_weighting)

            if 'pa' in outputs_lc:
                # ;Build pitch angle spectrogram
                out_pad_y[index, :], out_pad[index, :] = erg_pgs_make_theta_spec(clean_data, colatitude=True, resolution=regrid[1], no_ang_weighting=no_ang_weighting)

            if 'fac_energy' in outputs_lc:
                out_fac_energy_y[index, :], out_fac_energy[index, :] = erg_pgs_make_e_spec(clean_data)


    if 'energy' in outputs_lc:
        output_tplot_name = in_tvarname+'_energy' + suffix
        erg_pgs_make_tplot(output_tplot_name, x=times_array, y=out_energy_y, z=out_energy, units=units, ylog=True, ytitle=dist['data_name'] + ' \\ energy (eV)',relativistic=relativistic)
        out_vars.append(output_tplot_name)
    if 'phi' in outputs_lc:
        output_tplot_name = in_tvarname+'_phi' + suffix
        erg_pgs_make_tplot(output_tplot_name, x=times_array, y=out_phi_y, z=out_phi, units=units, ylog=False, ytitle=dist['data_name'] + ' \\ phi (deg)',relativistic=relativistic)
        out_vars.append(output_tplot_name)

    #  ;;Pitch Angle Spectrograms
    if 'pa' in outputs_lc:
        output_tplot_name = in_tvarname+'_pa' + suffix
        erg_pgs_make_tplot(output_tplot_name, x=times_array, y=out_pad_y, z=out_pad, units=units, ylog=False, ytitle=dist['data_name'] + ' \\ PA (deg)',relativistic=relativistic)
        out_vars.append(output_tplot_name)


    if 'fac_energy' in outputs_lc:

        output_tplot_name = in_tvarname+'_energy_mag' + suffix
        erg_pgs_make_tplot(output_tplot_name, x=times_array, y=out_fac_energy_y, z=out_fac_energy, units=units, ylog=True, ytitle=dist['data_name'] + ' \\ energy (eV)',relativistic=relativistic)
        out_vars.append(output_tplot_name)

    return out_vars



import logging
from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot

from .config import CONFIG


def load(
    trange=["2014-10-18", "2014-10-19"],
    instrument="mag",
    datatype="",
    level="l2",
    suffix="",
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    time_clip=False,
):
    """
    Load MAVEN data files for a given time range and instrument.

    Parameters
    ----------
    trange : list of str, optional
        Time range in the format ['start_date', 'end_date'].
        Defaults to ['2014-10-18', '2014-10-19'].
    instrument : str, optional
        Instrument name. Defaults to 'mag'.
        Valid values are: 'mag', 'swea', 'swia', 'static', 'sep', 'kp' (or 'insitu').
    datatype : str, optional
        Data type, depends on instrument.
        Valid options (default is the first value in the list):
            mag (1): 'sunstate-1sec'
            swea (5): 'arc3d', 'arcpad', 'svy3d', 'svypad', 'svyspec'
            swia (6): 'onboardsvyspec', 'onboardsvymom', 'finesvy3d', 'finearc3d', 'coarsesvy3d', 'coarsearc3d'
            static (20): 'c0-64e2m', 'c2-32e32m', 'c4-4e64m', 'c6-32e64m', 'c8-32e16d', 'ca-16e4d16a', 'cc-32e8d32m',
                'cd-32e8d32m', 'ce-16e4d16a16m', 'cf-16e4d16a16m', 'd0-32e4d16a8m', 'd1-32e4d16a8m', 'd4-4d16a2m',
                'd6-events', 'd7-fsthkp', 'd8-12r1e', 'd9-12r64e', 'da-1r64e', 'db-1024tof', 'hkp'
            sep (2): 's1-cal-svy-full', 's2-cal-svy-full'
            kp or insitu (1): 'kp-4sec'
    level : str, optional
        Data level. Defaults to 'l2'.
    suffix : str, optional
        Suffix to append to variable names. Defaults to ''.
    get_support_data : bool, optional
        Whether to retrieve support data. Defaults to False.
        By default, only data with "VAR_TYPE=data" will be loaded into tplot.
        If set to True, "VAR_TYPE=support_data" will also be loaded into tplot.
    varformat : str, optional
        Variable format. Defaults to None.
    varnames : list of str, optional
        List of variable names to load. Defaults to [].
    downloadonly : bool, optional
        Whether to only download the files without loading them into tplot. Defaults to False.
    notplot : bool, optional
        Return the data in hash tables instead of creating tplot variables. Defaults to False.
    no_update : bool, optional
        Whether to skip updating the local data directory. Defaults to False.
    time_clip : bool, optional
        Whether to clip the loaded data to the specified time range. Defaults to False.

    Returns
    -------
    list of str
        List of loaded data variables or files downloaded.
    """
    if instrument == "mag":
        # https://spdf.gsfc.nasa.gov/pub/data/maven/mag/l2/
        # for mag, only valid options are (1): level=l2, datatype=sunstate-1sec
        pathformat = f"mag/l2/sunstate-1sec/cdfs/%Y/%m/mvn_mag_l2-sunstate-1sec_%Y%m%d_v??_r??.cdf"

    elif instrument == "swea":
        # https://spdf.gsfc.nasa.gov/pub/data/maven/swea/l2/
        # for swea, valid options are (5): level=l2, datatype=arc3d, arcpad, svy3d, svypad, svyspec
        if datatype == "":
            datatype = "arc3d"  # default
        elif datatype not in ["arc3d", "arcpad", "svy3d", "svypad", "svyspec"]:
            logging.error(
                f"Invalid datatype: {datatype}. Valid options are: arc3d, arcpad, svy3d, svypad, svyspec"
            )
            return
        pathformat = (
            f"swea/l2/{datatype}/%Y/%m/mvn_swe_l2_{datatype}_%Y%m%d_v??_r??.cdf"
        )

    elif instrument == "swia":
        # https://spdf.gsfc.nasa.gov/pub/data/maven/swia/l2/
        # for swia, valid options are (6): level=l2, datatype=onboardsvyspec, onboardsvymom, finesvy3d, finearc3d, coarsesvy3d, coarsearc3d
        if datatype == "":
            datatype = "onboardsvyspec"  # default
        elif datatype not in [
            "onboardsvyspec",
            "onboardsvymom",
            "finesvy3d",
            "finearc3d",
            "coarsesvy3d",
            "coarsearc3d",
        ]:
            logging.error(
                f"Invalid datatype: {datatype}. Valid options are: onboardsvyspec, onboardsvymom, finesvy3d, finearc3d, coarsesvy3d, coarsearc3d"
            )
            return
        pathformat = (
            f"swia/l2/{datatype}/%Y/%m/mvn_swi_l2_{datatype}_%Y%m%d_v??_r??.cdf"
        )

    elif instrument == "static":
        # https://spdf.gsfc.nasa.gov/pub/data/maven/static/l2/
        # for static, valid options are: level=l2, datatype=c0-64e2m, c2-32e32m, c4-4e64m, c6-32e64m, c8-32e16d, ca-16e4d16a,
        #    cc-32e8d32m, cd-32e8d32m, ce-16e4d16a16m, cf-16e4d16a16m, d0-32e4d16a8m, d1-32e4d16a8m, d4-4d16a2m, d6-events,
        #    d7-fsthkp, d8-12r1e, d9-12r64e, da-1r64e, db-1024tof, hkp
        static_dict = {
            "c0-64e2m": "c0-64e2m",
            "c2-32e32m": "c2-32e32m",
            "c4-4e64m": "c4-4e64m",
            "c6-32e64m": "c6-32e64m",
            "c8-32e16d": "c8-32e16d",
            "ca-16e4d16a": "ca-16e4d16a",
            "cc-32e8d32m": "cc-32e8d32m",
            "cd-32e8d32m": "cd-32e8d32m",
            "ce-16e4d16a16m": "ce-16e4d16a16m",
            "cf-16e4d16a16m": "cf-16e4d16a16m",
            "d0-32e4d16a8m": "d0-32e4d16a8m",
            "d1-32e4d16a8m": "d1-32e4d16a8m",
            "d4-4d16a2m": "d4-4d16a2m",
            "d6-events": "d6-events",
            "d7-fsthkp": "d7-fsthkp",
            "d8-12r1e": "d8-12r1e",
            "d9-12r64e": "d9-12r64e",
            "da-1r64e": "da-1r64e",
            "db-1024tof": "db-1024tof",
            "hkp": "2a-hkp",
        }
        if datatype == "":
            datatype = "c0-64e2m"  # default
        elif datatype not in static_dict.keys():
            logging.error(
                f"Invalid datatype: {datatype}. Valid options are: c0-64e2m, c2-32e32m, c4-4e64m, c6-32e64m, c8-32e16d, ca-16e4d16a, "
                "cc-32e8d32m, cd-32e8d32m, ce-16e4d16a16m, cf-16e4d16a16m, d0-32e4d16a8m, d1-32e4d16a8m, d4-4d16a2m, d6-events, "
                "d7-fsthkp, d8-12r1e, d9-12r64e, da-1r64e, db-1024tof, hkp"
            )
            return
        
        datatypefile = static_dict[datatype]
        pathformat = (
            f"static/l2/{datatype}/%Y/%m/mvn_sta_l2_{datatypefile}_%Y%m%d_v??_r??.cdf"
        )

    elif instrument == "sep":
        # https://spdf.gsfc.nasa.gov/pub/data/maven/sep/l2/
        # for sep, valid options are (2): level=l2, datatype=s1-cal-svy-full, s2-cal-svy-full
        if datatype == "":
            datatype = "s1-cal-svy-full"  # default
        elif datatype not in ["s1-cal-svy-full", "s2-cal-svy-full"]:
            logging.error(
                f"Invalid datatype: {datatype}. Valid options are: s1-cal-svy-full, s2-cal-svy-full"
            )
            return
        pathformat = f"sep/l2/{datatype}/%Y/%m/mvn_sep_l2_{datatype}_%Y%m%d_v??_r??.cdf"

    elif instrument == "kp" or instrument == "insitu":
        # https://spdf.gsfc.nasa.gov/pub/data/maven/insitu/kp-4sec/cdfs/
        # for kp, only valid options are (1): level=l2, datatype=kp-4sec
        pathformat = f"insitu/kp-4sec/cdfs/%Y/%m/mvn_insitu_kp-4sec_%Y%m%d_v??_r??.cdf"

    else:
        logging.error(
            f"Invalid instrument: {instrument}. Valid options for MAVEN SPDF downloads are: mag, swea, swia, static, sep, kp"
        )
        return

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(
        remote_file=remote_names,
        remote_path=CONFIG["remote_data_dir"],
        local_path=CONFIG["local_data_dir"],
        no_download=no_update,
    )
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(
        out_files,
        suffix=suffix,
        get_support_data=get_support_data,
        varformat=varformat,
        varnames=varnames,
        notplot=notplot,
    )

    if tvars is None or notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix="")

    return tvars



from .load import load
from pyspedas.utilities.datasets import find_datasets


def datasets(instrument=None, label=True):
    """
    Query SPDF for datasets available for a given Cluster instrument

    Parameters
    ----------
        instrument : str
            Instrument to use in query. Valid options: 'ASP','CIS','DWP','EDI','EFW','FGM','PEA','RAP','STA','WBD','WHI'
            Default: None

        label: bool
            If True, print both the dataset name and label; otherwise print only the dataset name.


    Examples
    --------
    >>> import pyspedas
    >>> pyspedas.projects.cluster.datasets('FGM')
    """
    return find_datasets(mission='Cluster', instrument=instrument, label=label)



import logging
import re
import os
from . import download_files_utilities as utils
from .file_regex import maven_kp_l2_regex
import numpy as np
import collections

from pyspedas.utilities.download import is_fsspec_uri
import fsspec

def read_iuvs_file(file):
    """
    Read an IUVS file and return a dictionary containing the data.

    Parameters:
        file (str): The path to the IUVS file to be read.

    Returns:
        dict: A dictionary containing the data read from the IUVS file.
    """
    iuvs_dict = {}
    periapse_num = 0
    occ_num = 0
    if is_fsspec_uri(file):
        protocol, path = file.split("://")
        fs = fsspec.filesystem(protocol)

        fileobj = fs.open(file, "r")
    else:
        fileobj = open(file, "r")
    with fileobj as f:
        line = f.readline()
        while line != "":
            if line.startswith("*"):
                # Read the header
                line = f.readline()
                obs_mode = line[19 : len(line) - 1].strip()

                header = {}
                f.readline()
                line = f.readline()
                header["time_start"] = line[19 : len(line) - 1].strip()
                line = f.readline()
                header["time_stop"] = line[19 : len(line) - 1].strip()
                line = f.readline()
                if obs_mode == "OCCULTATION":
                    header["target_name"] = line[19 : len(line) - 1].strip()
                    line = f.readline()
                header["sza"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["local_time"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["lat"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["lon"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["lat_mso"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["lon_mso"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["orbit_number"] = int(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["mars_season"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_geo_x"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_geo_y"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_geo_z"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_mso_x"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_mso_y"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_mso_z"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sun_geo_x"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sun_geo_y"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sun_geo_z"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sun_geo_lat"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sun_geo_lon"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sun_mso_lat"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sun_mso_lon"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["subsol_geo_lon"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["subsol_geo_lat"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_sza"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_local_time"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["sc_alt"] = float(line[19 : len(line) - 1].strip())
                line = f.readline()
                header["mars_sun_dist"] = float(line[19 : len(line) - 1].strip())

                if obs_mode == "PERIAPSE":
                    periapse_num += 1
                    line = f.readline()
                    n_alt_bins = int(line[19 : len(line) - 1].strip())
                    header["n_alt_bins"] = float(n_alt_bins)
                    line = f.readline()
                    n_alt_den_bins = int(line[19 : len(line) - 1].strip())
                    header["n_alt_den_bins"] = float(n_alt_den_bins)

                    iuvs_dict["periapse" + str(periapse_num)] = {}
                    iuvs_dict["periapse" + str(periapse_num)].update(header)

                    # Empty space
                    f.readline()

                    # Read the Temperature
                    line = f.readline()
                    temp_labels = line[19 : len(line) - 1].strip().split()
                    temperature = collections.OrderedDict((x, []) for x in temp_labels)
                    temperature_unc = collections.OrderedDict(
                        (x, []) for x in temp_labels
                    )
                    line = f.readline()
                    vals = line[20 : len(line) - 1].strip().split()
                    index = 0
                    for val in vals:
                        if val == "-9.9999990E+09":
                            val = float("nan")
                        else:
                            val = float(val)
                        temperature[list(temperature.keys())[index]].append(val)
                        index += 1
                    line = f.readline()
                    vals = line[20 : len(line) - 1].strip().split()
                    index = 0
                    for val in vals:
                        if val == "-9.9999990E+09":
                            val = float("nan")
                        else:
                            val = float(val)
                        temperature_unc[list(temperature_unc.keys())[index]].append(val)
                        index += 1
                    iuvs_dict["periapse" + str(periapse_num)][
                        "temperature"
                    ] = temperature
                    iuvs_dict["periapse" + str(periapse_num)][
                        "temperature_unc"
                    ] = temperature_unc

                    # Empty space
                    f.readline()

                    # Read the Scale Heights
                    line = f.readline()
                    scale_height_labels = line[19 : len(line) - 1].strip().split()
                    scale_height = collections.OrderedDict(
                        (x, []) for x in scale_height_labels
                    )
                    scale_height_unc = collections.OrderedDict(
                        (x, []) for x in scale_height_labels
                    )
                    line = f.readline()
                    vals = line[20 : len(line) - 1].strip().split()
                    index = 0
                    for val in vals:
                        if val == "-9.9999990E+09":
                            val = float("nan")
                        else:
                            val = float(val)
                        scale_height[list(scale_height.keys())[index]].append(val)
                        index += 1
                    line = f.readline()
                    vals = line[20 : len(line) - 1].strip().split()
                    index = 0
                    for val in vals:
                        if val == "-9.9999990E+09":
                            val = float("nan")
                        else:
                            val = float(val)
                        scale_height_unc[list(scale_height_unc.keys())[index]].append(
                            val
                        )
                        index += 1

                    iuvs_dict["periapse" + str(periapse_num)][
                        "scale_height"
                    ] = scale_height
                    iuvs_dict["periapse" + str(periapse_num)][
                        "scale_height_unc"
                    ] = scale_height_unc

                    # Empty space
                    f.readline()
                    f.readline()

                    # Read in the density
                    line = f.readline()
                    density_labels = line.strip().split()
                    density = collections.OrderedDict((x, []) for x in density_labels)
                    for i in range(0, n_alt_den_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == "-9.9999990E+09":
                                val = float("nan")
                            else:
                                val = float(val)
                            density[list(density.keys())[index]].append(val)
                            index += 1
                    iuvs_dict["periapse" + str(periapse_num)]["density"] = density

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the density systematic uncertainty
                    density_sys_unc = collections.OrderedDict(
                        (x, []) for x in density_labels
                    )
                    line = f.readline()
                    vals = line.strip().split()
                    index = 0
                    for val in vals:
                        if val == "-9.9999990E+09":
                            val = float("nan")
                        else:
                            val = float(val)
                        density_sys_unc[list(density.keys())[index + 1]].append(val)
                        index += 1

                    iuvs_dict["periapse" + str(periapse_num)][
                        "density_sys_unc"
                    ] = density_sys_unc

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the density uncertainty
                    density_unc = collections.OrderedDict(
                        (x, []) for x in density_labels
                    )
                    for i in range(0, n_alt_den_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == "-9.9999990E+09":
                                val = float("nan")
                            else:
                                val = float(val)
                            density_unc[list(density.keys())[index]].append(val)
                            index += 1
                    iuvs_dict["periapse" + str(periapse_num)][
                        "density_sys_unc"
                    ] = density_sys_unc

                    f.readline()
                    f.readline()

                    line = f.readline()
                    radiance_labels = line.strip().split()
                    if "Cameron" in radiance_labels:
                        radiance_labels.remove("Cameron")
                    radiance = collections.OrderedDict((x, []) for x in radiance_labels)
                    for i in range(0, n_alt_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == "-9.9999990E+09":
                                val = float("nan")
                            else:
                                val = float(val)
                            radiance[list(radiance.keys())[index]].append(val)
                            index += 1

                    iuvs_dict["periapse" + str(periapse_num)]["radiance"] = radiance

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the radiance systematic uncertainty
                    radiance_sys_unc = collections.OrderedDict(
                        (x, []) for x in radiance_labels
                    )
                    line = f.readline()
                    vals = line.strip().split()
                    index = 0
                    for val in vals:
                        if val == "-9.9999990E+09":
                            val = float("nan")
                        else:
                            val = float(val)
                        radiance_sys_unc[list(radiance.keys())[index + 1]].append(val)
                        index += 1

                    iuvs_dict["periapse" + str(periapse_num)][
                        "radiance_sys_unc"
                    ] = radiance_sys_unc

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the radiance uncertainty
                    radiance_unc = collections.OrderedDict(
                        (x, []) for x in radiance_labels
                    )
                    for i in range(0, n_alt_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == "-9.9999990E+09":
                                val = float("nan")
                            else:
                                val = float(val)
                            radiance_unc[list(radiance.keys())[index]].append(val)
                            index += 1

                    iuvs_dict["periapse" + str(periapse_num)][
                        "radiance_unc"
                    ] = radiance_unc

                elif obs_mode == "OCCULTATION":
                    occ_num += 1
                    line = f.readline()
                    n_alt_den_bins = int(line[19 : len(line) - 1].strip())
                    header["n_alt_den_bins"] = float(n_alt_den_bins)

                    iuvs_dict["occultation" + str(occ_num)] = {}
                    iuvs_dict["occultation" + str(occ_num)].update(header)

                    # Empty space
                    f.readline()

                    # Read the Scale Heights
                    line = f.readline()
                    scale_height_labels = line[19 : len(line) - 1].strip().split()
                    scale_height = collections.OrderedDict(
                        (x, []) for x in scale_height_labels
                    )
                    scale_height_unc = collections.OrderedDict(
                        (x, []) for x in scale_height_labels
                    )
                    line = f.readline()
                    vals = line[20 : len(line) - 1].strip().split()
                    index = 0
                    for val in vals:
                        if val == "-9.9999990E+09":
                            val = float("nan")
                        else:
                            val = float(val)
                        scale_height[list(scale_height.keys())[index]].append(val)
                        index += 1
                    line = f.readline()
                    vals = line[20 : len(line) - 1].strip().split()
                    index = 0
                    for val in vals:
                        if val == "-9.9999990E+09":
                            val = float("nan")
                        else:
                            val = float(val)
                        scale_height_unc[list(scale_height_unc.keys())[index]].append(
                            val
                        )
                        index += 1

                    iuvs_dict["occultation" + str(occ_num)][
                        "scale_height"
                    ] = scale_height
                    iuvs_dict["occultation" + str(occ_num)][
                        "scale_height_unc"
                    ] = scale_height_unc

                    # Empty space
                    f.readline()
                    f.readline()

                    # Read in the retrieval
                    line = f.readline()
                    retrieval_labels = line.strip().split()
                    retrieval = collections.OrderedDict(
                        (x, []) for x in retrieval_labels
                    )
                    for i in range(0, n_alt_den_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == "-9.9999990E+09":
                                val = float("nan")
                            else:
                                val = float(val)
                            retrieval[list(retrieval.keys())[index]].append(val)
                            index += 1
                    iuvs_dict["occultation" + str(occ_num)]["retrieval"] = retrieval

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the retrieval systematic uncertainty
                    retrieval_sys_unc = collections.OrderedDict(
                        (x, []) for x in retrieval_labels
                    )
                    line = f.readline()
                    vals = line.strip().split()
                    index = 0
                    for val in vals:
                        if val == "-9.9999990E+09":
                            val = float("nan")
                        else:
                            val = float(val)
                        retrieval_sys_unc[list(retrieval.keys())[index + 1]].append(val)
                        index += 1

                    iuvs_dict["occultation" + str(occ_num)][
                        "retrieval_sys_unc"
                    ] = retrieval_sys_unc

                    # Not needed lines
                    f.readline()
                    f.readline()
                    f.readline()

                    # Read in the retrieval uncertainty
                    retrieval_unc = collections.OrderedDict(
                        (x, []) for x in retrieval_labels
                    )
                    for i in range(0, n_alt_den_bins):
                        line = f.readline()
                        vals = line.strip().split()
                        index = 0
                        for val in vals:
                            if val == "-9.9999990E+09":
                                val = float("nan")
                            else:
                                val = float(val)
                            retrieval_unc[list(retrieval.keys())[index]].append(val)
                            index += 1
                    iuvs_dict["occultation" + str(occ_num)][
                        "retrieval_sys_unc"
                    ] = retrieval_sys_unc

            line = f.readline()

    return iuvs_dict


